from .base import Guest


class OpenSUSEGuest(Guest):
    """ This guest can provision openSUSE containers. """

    name = 'opensuse'

    def install_packages(self, packages):
        self.run(['zypper', '--non-interactive', 'install', ] + packages)



from .base import Guest


class ArchLinuxGuest(Guest):
    """ This guest can provision ArchLinux containers. """

    name = 'arch'

    def install_packages(self, packages):
        self.run(['pacman', '-S', '--noconfirm'] + packages)



import logging
import os
from pathlib import Path

import yaml
from dotenv.main import dotenv_values
from voluptuous.error import Invalid

from . import constants
from .exceptions import (ConfigFileInterpolationError, ConfigFileNotFoundError,
                         ConfigFileValidationError)
from .interpolation import interpolate_variables
from .schema import schema


logger = logging.getLogger(__name__)


class Config:
    """ Holds the configuration of a LXDock project. """

    def __init__(self, homedir, filename):
        self.homedir = homedir
        self.filename = filename
        self.containers = []
        self.provisioning_steps = []
        self._dict = {}

    def __contains__(self, key):
        return key in self._dict

    def __getitem__(self, key):
        return self._dict[key]

    @classmethod
    def from_base_dir(cls, base_dir='.'):
        """ Returns a Config instance using a base directory. """
        base_dir_path = Path(os.path.abspath(base_dir))
        candidate_paths = [base_dir_path, ] + list(base_dir_path.parents)
        existing_config_paths = []
        for candidate_path in candidate_paths:
            existing_config_paths = [
                os.path.join(str(candidate_path), filename)
                for filename in constants.ALLOWED_FILENAMES
                if os.path.exists(os.path.join(str(candidate_path), filename))]
            if existing_config_paths:
                break

        if not existing_config_paths:
            raise ConfigFileNotFoundError(
                'Unable to find a suitable configuration file in this directory. '
                'Are you in the right directory?\n'
                'The supported filenames are: {}'.format(', '.join(constants.ALLOWED_FILENAMES))
            )

        config_dirname, config_filename = os.path.split(existing_config_paths[0])
        if len(existing_config_paths) > 1:
            logger.warn('Multiple config files were found: {0}'.format(
                ', '.join([os.path.split(p)[1] for p in existing_config_paths])))
            logger.warn('Using: {0}'.format(config_filename))

        # Initializes the config instance.
        config = cls(config_dirname, config_filename)

        # Loads the YML.
        config.load()

        # Validates the content of the configuration. We chdir into the home directory of the
        # project in order to ensure that IsFile/IsDir validators keep working properly if the
        # config is initialized from a subfolder of the project.
        cwd = os.getcwd()
        os.chdir(config.homedir)
        try:
            schema(config._dict)
        except Invalid as e:
            # Formats the voluptuous error
            path = ' @ %s' % '.'.join(map(str, e.path)) if e.path else ''
            msg = 'The LXDock file is invalid because: {0}'.format(e.msg + path)
            raise ConfigFileValidationError(msg)
        finally:
            os.chdir(cwd)

        # Performs variable substitution / interpolation in the configuration values.
        config.interpolate()

        config.extract_config_from_dict()

        return config

    def interpolate(self):
        """ Interpolates the considered config.

        In order to perform this variable substitution LXDock will use environment variables plus
        some other pre-defined variables. The later will be injected in the context used to perform
        this interpolation.
        """
        # Builds a dictionary of variables that will be used to perform variable substitution in the
        # config values.
        mapping = {}
        mapping.update(os.environ)

        # Fetches variables that could be defined in a .env file and inserts them in the mapping.
        env_filepath = os.path.join(self.homedir, '.env')
        if os.path.exists(env_filepath):
            mapping.update(dotenv_values(env_filepath))

        # Inserts LXDock special variables into the final mapping.
        mapping.update({
            # The absolute path toward the home directory of the LXDock project is injected into
            # the mapping used to perform variable substitution in the initial config.
            # This can be useful to reference paths that are relative to the project's root
            # directory in LXDock file options such as inline commands.
            'LXDOCK_YML_DIR': self.homedir,
        })

        try:
            self._dict = interpolate_variables(self._dict, mapping)
        except KeyError as e:
            raise ConfigFileInterpolationError(
                'Variable substitution failed when parsing the LXDock file. The failing variable '
                'name is: ${}'.format(e.args[0]))

    def load(self):
        """ Loads the YML configuration file and store it inside the `_dict` dictionary. """
        self._dict = self._load_yml()

    def extract_config_from_dict(self):
        """ Take config from the _dict and place them into ready-to-use config in `self`. """
        containers = [ContainerConfig(self._get_container_config_dict(cdict))
                      for cdict in self._dict.get('containers', [])]
        # If we cannot consider multiple containers, we just pass the full dictionary to initialize
        # the `ContainerConfig` instance.
        if not len(containers):
            unique_container_config = ContainerConfig(self._dict)
            # We don't inherit global provisioning in specific container config.
            if 'provisioning' in unique_container_config:
                del unique_container_config['provisioning']

            # We associate a specific name to the unique container if it wasn't explicitly defined
            # under the 'containers' section of the LXDock file.
            unique_container_config['name'] = 'default'
            containers = [unique_container_config, ]

        self.containers.extend(containers)

        self.provisioning_steps = self._dict.get('provisioning', [])

    def serialize(self):
        """ Returns the configuration as a string. """
        return yaml.dump(self._dict, default_flow_style=False)

    def _get_container_config_dict(self, container_dict):
        """ Returns a dictionary containing the container's configuration.

        The dictionary that is returned contains the container's configuration and the global config
        values. These global values can be defined outside of the scope of the container's config
        and can be used by each container.
        """
        container_config = dict(self._dict)

        # We don't inherit global provisioning in specific container config.
        if 'provisioning' in container_config:
            del container_config['provisioning']

        # If both global and container scope contains lxc_config, merge them
        lxck = 'lxc_config'
        if lxck in container_config and lxck in container_dict:
            container_dict = dict(container_dict)
            container_config[lxck].update(container_dict[lxck])
            del container_dict[lxck]

        container_config.update(container_dict)
        del container_config['containers']
        return container_config

    def _load_yml(self):
        """ Loads the YML configuration file. """
        with open(os.path.join(self.homedir, self.filename), 'r') as fp:
            return yaml.safe_load(fp)
        # TODO: handle potential errors here


class ContainerConfig(dict):
    """ Holds the specific configuration of a container. """



"""
    LXDock test tools
    =================
    This package provide test tools such as testcases, fixtures or helpers that can be used to setup
    or run tests.
"""

from .fakes import *  # noqa
from .fixtures import *  # noqa
from .testcases import *  # noqa



from .base import Guest


class OracleLinuxGuest(Guest):
    """ This guest can provision Oracle Linux containers. """

    name = 'ol'

    def install_packages(self, packages):
        self.run(['yum', '-y', 'install'] + packages)



import os
import pathlib
import tempfile
import unittest.mock

import pytest
from pylxd.exceptions import NotFound

from lxdock.guests import Guest
from lxdock.guests.base import InvalidGuest
from lxdock.test import FakeContainer


class TestGuest:
    def test_subclasses_cannot_be_registered_if_they_do_not_provide_a_guest_name(self):
        with pytest.raises(InvalidGuest):
            class InvalidDummyGuest(Guest):
                name = None

    def test_can_detect_os_or_distribution(self):
        class DummyGuest(Guest):
            name = 'dummy'
        lxd_container_1 = unittest.mock.Mock()
        lxd_container_1.files.get.return_value = 'ID=dummy'
        lxd_container_2 = unittest.mock.Mock()
        lxd_container_2.files.get.return_value = 'ID=unknown'
        lxd_container_3 = unittest.mock.Mock()
        lxd_container_3.files.get.side_effect = NotFound(response=unittest.mock.Mock())
        assert DummyGuest.detect(lxd_container_1)
        assert not DummyGuest.detect(lxd_container_2)
        assert not DummyGuest.detect(lxd_container_3)

    def test_can_add_ssh_pubkey_to_root_authorized_keys(self):
        class DummyGuest(Guest):
            name = 'dummy'
        guest = DummyGuest(FakeContainer())
        guest.add_ssh_pubkey_to_root_authorized_keys('pubkey')
        assert guest.lxd_container.execute.call_count == 1
        assert guest.lxd_container.execute.call_args[0] == (['mkdir', '-p', '/root/.ssh'], )
        assert guest.lxd_container.files.put.call_count == 1
        assert guest.lxd_container.files.put.call_args[0] == \
            ('/root/.ssh/authorized_keys', 'pubkey', )

    def test_can_create_a_user_with_a_default_home_directory(self):
        class DummyGuest(Guest):
            name = 'dummy'
        guest = DummyGuest(FakeContainer())
        guest.create_user('usertest')
        assert guest.lxd_container.execute.call_count == 1
        assert guest.lxd_container.execute.call_args[0] == \
            (['useradd', '--create-home', 'usertest'], )

    def test_can_create_a_user_with_a_custom_home_directory(self):
        class DummyGuest(Guest):
            name = 'dummy'
        guest = DummyGuest(FakeContainer())
        guest.create_user('usertest', home='/opt/usertest')
        assert guest.lxd_container.execute.call_count == 1
        assert guest.lxd_container.execute.call_args[0] == \
            (['useradd', '--create-home', '--home-dir', '/opt/usertest', 'usertest'], )

    def test_can_create_a_user_with_a_custom_password(self):
        class DummyGuest(Guest):
            name = 'dummy'
        password = '$6$cGzZBkDjOhGW$6C9wwqQteFEY4lQ6ZJBggE568SLSS7bIMKexwOD' \
                   '39mJQrJcZ5vIKJVIfwsKOZajhbPw0.Zqd0jU2NDLAnp9J/1'
        guest = DummyGuest(FakeContainer())
        guest.create_user('usertest', password=password)
        assert guest.lxd_container.execute.call_count == 1
        assert guest.lxd_container.execute.call_args[0] == \
            (['useradd', '--create-home', '-p', password, 'usertest'], )

    def test_can_copy_file(self):
        class DummyGuest(Guest):
            name = 'dummy'
        guest = DummyGuest(FakeContainer())
        guest.lxd_container.files.put.return_value = True
        with tempfile.NamedTemporaryFile() as f:
            f.write(b'dummy file')
            f.flush()
            guest.copy_file(pathlib.Path(f.name), pathlib.PurePath('/a/b/c'))
        assert guest.lxd_container.execute.call_count == 1
        assert guest.lxd_container.execute.call_args[0] == (['mkdir', '-p', '/a/b'], )
        assert guest.lxd_container.files.put.call_count == 1
        assert guest.lxd_container.files.put.call_args[0] == ('/a/b/c', b'dummy file')

    @unittest.mock.patch('tarfile.TarFile.add')
    @unittest.mock.patch('tarfile.TarFile.close')
    def test_can_copy_directory(self, mock_close, mock_add):
        class DummyGuest(Guest):
            name = 'dummy'
        guest = DummyGuest(FakeContainer())
        guest.lxd_container.files.put.return_value = True
        with tempfile.TemporaryDirectory() as d:
            os.mkdir('{}/d1'.format(d))
            os.mkdir('{}/d1/d2'.format(d))
            with open('{}/d1/f1'.format(d), 'wb') as f1:
                f1.write(b'dummy f1')
            with open('{}/d1/d2/f2'.format(d), 'wb') as f2:
                f2.write(b'dummy f2')
            with open('{}/f3'.format(d), 'wb') as f3:
                f3.write(b'dummy f3')
            guest.copy_directory(pathlib.Path(d), pathlib.PurePosixPath('/a/b/c'))

        assert mock_add.call_count == 1
        assert mock_add.call_args[0][0] == str(pathlib.Path(d))
        assert mock_add.call_args[1]['arcname'] == '.'

        assert mock_close.call_count == 1

        assert guest.lxd_container.execute.call_count == 4
        assert guest.lxd_container.execute.call_args_list[0][0] == (['mkdir', '-p', '/a/b/c'], )
        assert guest.lxd_container.execute.call_args_list[1][0] == ([
            'mkdir', '-p', str(pathlib.PurePosixPath(guest._guest_temporary_tar_path).parent)], )
        assert guest.lxd_container.execute.call_args_list[2][0] == ([
            'tar', '-xf', guest._guest_temporary_tar_path, '-C', '/a/b/c'], )
        assert guest.lxd_container.execute.call_args_list[3][0] == ([
            'rm', '-f', guest._guest_temporary_tar_path], )

        assert guest.lxd_container.files.put.call_count == 1
        assert guest.lxd_container.files.put.call_args[0][0] == guest._guest_temporary_tar_path



import unittest.mock

import pytest

from lxdock.guests import DebianGuest
from lxdock.provisioners import Provisioner
from lxdock.provisioners.base import InvalidProvisioner
from lxdock.test import FakeContainer


class TestProvisioner:
    def test_subclasses_cannot_be_registered_if_they_do_not_provide_a_provisioner_name(self):
        with pytest.raises(InvalidProvisioner):
            class InvalidDummyProvisioner(Provisioner):
                name = None
                schema = {}

    def test_subclasses_cannot_be_registered_if_they_do_not_provide_a_schema_dict(self):
        with pytest.raises(InvalidProvisioner):
            class InvalidDummyProvisioner(Provisioner):
                name = 'myprovisioner'
                schema = None

    def test_trigger_packages_installation_on_the_guest_if_the_related_attr_is_defined(self):
        class DummyProvisioner(Provisioner):
            name = 'myprovisioner'
            schema = {'test': 'test', }

            guest_required_packages_debian = ['test01', 'test02', ]

        container = FakeContainer()
        lxd_container = container._container
        lxd_container.execute.return_value = ('ok', 'ok', '')
        host = unittest.mock.Mock()
        guest = DebianGuest(container)
        provisioner = DummyProvisioner('./', host, [guest], {})
        provisioner.setup()
        assert lxd_container.execute.call_count == 2
        assert lxd_container.execute.call_args_list[0][0] == \
            (['apt-get', 'update'], )
        assert lxd_container.execute.call_args_list[1][0] == \
            (['apt-get', 'install', '-y', 'test01', 'test02', ], )

    def test_trigger_specific_setup_on_the_guest_if_the_related_method_is_defined(self):
        class DummyProvisioner(Provisioner):
            name = 'myprovisioner'
            schema = {'test': 'test', }
            called = False

            def setup_guest_debian(self, guest):
                self.called = True

        container = FakeContainer()
        lxd_container = container._container
        lxd_container.execute.return_value = ('ok', 'ok', '')
        host = unittest.mock.Mock()
        guest = DebianGuest(container)
        provisioner = DummyProvisioner('./', host, [guest], {})
        provisioner.setup()
        assert provisioner.called



import re
from string import Template


class ConfigTemplate(Template):
    """ Like string.Template except that only variables in the format
    ${var} are supported but not the unbraced format of $var.

    The stricter format allows $ characters to be used more easily in
    the lxdock.yml file without having to escape them.
    """
    pattern = r"""
    %(delim)s(?:
      {(?P<braced>%(id)s)}  |  # delimiter and a braced identifier
      {(?P<named>%(id)s)}   |  # delimiter and a Python identifier (not needed)
      (?P<escaped>)         |  # Escape sequence (not needed)
      (?P<invalid>)            # Other ill-formed delimiter exprs
    )
    """ % dict(delim=re.escape(Template.delimiter), id=Template.idpattern)


def interpolate_variables(config_dict, mapping):
    """ Performs variable interpolation in the considered config.

    :param config_dict: dictionary containing the configuration values
    :param mapping: dictionary containing the available variables
    :type config_dict: dict
    :type mapping: dict
    :return: dictionary containing the interpolated config
    :rtype: dict
    """
    def interpolate(value):
        if isinstance(value, str):
            return ConfigTemplate(value).substitute(**mapping)
        elif isinstance(value, dict):
            return {k: interpolate(v) for k, v in value.items()}
        elif isinstance(value, (list, tuple)):
            return [interpolate(v) for v in value]
        return value

    return {k: interpolate(v) for k, v in config_dict.items()}



from lxdock.guests import GentooGuest
from lxdock.test import FakeContainer


class TestGentooGuest:
    def test_should_install_packages_if_not_installed(self):
        guest = GentooGuest(FakeContainer())
        # Retcode 1: "No installed packages matching {keyword}"
        guest.lxd_container.execute.return_value = (1, 'ok', '')
        guest.install_packages(['python', 'openssh', ])
        assert guest.lxd_container.execute.call_count == 5
        assert guest.lxd_container.execute.call_args_list[0][0] == \
            (['emerge', 'app-portage/gentoolkit'], )
        for i, p in enumerate(['python', 'openssh', ]):
            assert guest.lxd_container.execute.call_args_list[2 * i + 1][0] == \
                (['equery', 'list', p], )
            assert guest.lxd_container.execute.call_args_list[2 * i + 2][0] == (['emerge', p], )

    def test_should_not_reinstall_packages_if_already_installed(self):
        guest = GentooGuest(FakeContainer())
        # Retcode 0: the package has been found locally
        guest.lxd_container.execute.return_value = (0, 'ok', '')
        guest.install_packages(['python', 'openssh', ])
        assert guest.lxd_container.execute.call_count == 3
        assert guest.lxd_container.execute.call_args_list[0][0] == \
            (['emerge', 'app-portage/gentoolkit'], )
        for i, p in enumerate(['python', 'openssh', ]):
            assert guest.lxd_container.execute.call_args_list[i + 1][0] == (['equery', 'list', p], )



import unittest.mock

import pytest
from voluptuous import All, Any, Boolean, Coerce, Length
from voluptuous.error import Invalid

from lxdock.conf.schema import get_schema


class MockProvisioner1(object):
    name = 'mp1'
    schema = {'a': str, 'b': Coerce(int)}


class MockProvisioner2(object):
    name = 'mp2'
    schema = All({'a': str}, {'a': Length(min=5, max=5)})


class MockProvisioner3(object):
    name = 'mp3'
    schema = Any({'a': str}, {'b': Boolean()})


class TestValidateProvisionerSchemas:
    @unittest.mock.patch('lxdock.conf.schema.Provisioner')
    def test_can_validate_and_coerce_multiple_provisioner_schemas(self, mock_Provisioner):
        mock_Provisioner.provisioners = {
            'mp1': MockProvisioner1,
            'mp2': MockProvisioner2,
            'mp3': MockProvisioner3}
        schema = get_schema()
        validated = schema({
            'name': 'dummy-test',
            'provisioning': [{
                'type': 'mp1',
                'a': 'dummy',
                'b': '16'
            }, {
                'type': 'mp2',
                'a': 'dummy',
            }, {
                'type': 'mp3',
                'b': 'yes'
            }]
        })
        assert validated == {
            'name': 'dummy-test',
            'provisioning': [{
                'type': 'mp1',
                'a': 'dummy',
                'b': 16  # Check Coerce
            }, {
                'type': 'mp2',
                'a': 'dummy',
            }, {
                'type': 'mp3',
                'b': True  # Check Boolean
            }]
        }

    @unittest.mock.patch('lxdock.conf.schema.Provisioner')
    def test_raise_invalid_if_provisioner_schema_is_not_satisfied(self, mock_Provisioner):
        mock_Provisioner.provisioners = {
            'mp1': MockProvisioner1,
            'mp2': MockProvisioner2,
            'mp3': MockProvisioner3}
        schema = get_schema()
        with pytest.raises(Invalid) as e:
            schema({
                'name': 'dummy-test',
                'provisioning': [{
                    'type': 'mp1',
                    'a': 'dummy',
                    'b': '16'
                }, {
                    'type': 'mp2',
                    'a': 'dummydummy',  # Exceeds Length(min=5, max=5)
                }, {
                    'type': 'mp3',
                    'b': 'yes'
                }]
            })
        assert "['provisioning'][1]['a']" in str(e)



from lxdock.guests import AlpineGuest
from lxdock.test import FakeContainer


class TestAlpineGuest:
    def test_can_install_packages(self):
        guest = AlpineGuest(FakeContainer())
        guest.install_packages(['python', 'openssh', ])
        assert guest.lxd_container.execute.call_count == 2
        assert guest.lxd_container.execute.call_args_list[0][0] == (['apk', 'update'], )
        assert guest.lxd_container.execute.call_args_list[1][0] == \
            (['apk', 'add', 'python', 'openssh'], )



"""
    Guest components
    ================
    This package contains guest definitions. Guests are used to perform operations on the "guest"
    side (the containers).
"""

from .alpine import *  # noqa
from .archlinux import *  # noqa
from .base import *  # noqa
from .centos import *  # noqa
from .debian import *  # noqa
from .fedora import *  # noqa
from .gentoo import *  # noqa
from .opensuse import *  # noqa
from .oracle import *  # noqa
from .ubuntu import *  # noqa



import unittest.mock

from lxdock.guests import DebianGuest
from lxdock.hosts import Host
from lxdock.provisioners import ShellProvisioner
from lxdock.test import FakeContainer


class TestShellProvisioner:
    @unittest.mock.patch('subprocess.Popen')
    def test_can_run_commands_on_the_host_side(self, mock_popen):
        host = Host()
        guest = DebianGuest(FakeContainer())
        provisioner = ShellProvisioner(
            './', host, [guest], {
                'inline': """touch f && echo "Here's the PATH" $PATH >> /tmp/test.txt""",
                'side': 'host', })
        provisioner.provision()
        assert mock_popen.call_args[0] == (
            """sh -c 'touch f && echo "Here'"'"'s the PATH" $PATH >> /tmp/test.txt'""", )

    def test_can_run_commands_on_the_guest_side(self):
        container = FakeContainer()
        lxd_container = container._container
        host = Host()
        guest = DebianGuest(container)
        cmd = """touch f && echo "Here's the PATH" $PATH >> /tmp/test.txt"""
        provisioner = ShellProvisioner(
            './', host, [guest], {'inline': cmd})
        provisioner.provision()
        assert lxd_container.execute.call_count == 1
        assert lxd_container.execute.call_args_list[0][0] == (['sh', '-c', cmd], )

    @unittest.mock.patch('subprocess.Popen')
    def test_can_run_a_script_on_the_host_side(self, mock_popen):
        host = Host()
        guest = DebianGuest(unittest.mock.Mock())
        provisioner = ShellProvisioner(
            './', host, [guest], {'script': 'test.sh', 'side': 'host', })
        provisioner.provision()
        assert mock_popen.call_args[0] == ('./test.sh', )

    @unittest.mock.patch('builtins.open')
    def test_can_run_a_script_on_the_guest_side(self, mock_open):
        container = FakeContainer()
        lxd_container = container._container
        host = Host()
        guest = DebianGuest(container)
        provisioner = ShellProvisioner(
            './', host, [guest], {'script': 'test.sh', })
        provisioner.provision()
        assert lxd_container.execute.call_count == 2
        assert lxd_container.execute.call_args_list[0][0] == (['chmod', '+x', '/tmp/test.sh', ], )
        assert lxd_container.execute.call_args_list[1][0] == (['/tmp/test.sh', ], )



from test.support import EnvironmentVarGuard

from lxdock.utils.lxd import get_lxd_dir


def test_get_lxd_helper_can_return_the_lxd_base_directory():
    env = EnvironmentVarGuard()
    with env:
        assert get_lxd_dir() == '/var/lib/lxd'
        env.set('LXD_DIR', '/my/test/lxd/')
        assert get_lxd_dir() == '/my/test/lxd/'



import os
import sys
import types
import unittest.mock

import pytest
from pylxd.exceptions import NotFound

from lxdock import constants
from lxdock.container import Container, must_be_created_and_running
from lxdock.exceptions import ContainerOperationFailed
from lxdock.test.testcases import LXDTestCase


THIS_DIR = os.path.join(os.path.dirname(__file__))


def test_must_be_created_and_running_decorator_works(persistent_container):
    @must_be_created_and_running
    def dummy_action(self):
        return 42
    persistent_container.dummy_action = types.MethodType(dummy_action, persistent_container)
    persistent_container.halt()
    assert persistent_container.dummy_action() is None
    persistent_container.up()
    assert persistent_container.dummy_action() == 42
    del persistent_container.dummy_action

    non_created_container_options = {
        'name': 'lxdock-nonexistingcontainer', 'image': 'alpine/3.6', 'mode': 'pull', }
    non_created_container = Container(
        'myproject', THIS_DIR, persistent_container.client, **non_created_container_options)
    non_created_container.dummy_action = types.MethodType(dummy_action, non_created_container)
    assert non_created_container.dummy_action() is None


class TestContainer(LXDTestCase):
    def test_can_set_up_a_container_that_does_not_exist(self):
        container_options = {
            'name': self.containername('newcontainer'), 'image': 'alpine/3.6', 'mode': 'pull', }
        container = Container('myproject', THIS_DIR, self.client, **container_options)
        container.up()
        assert container._container.status_code == constants.CONTAINER_RUNNING
        assert container._container.config['user.lxdock.made'] == '1'
        assert container._container.config['user.lxdock.homedir'] == THIS_DIR

    def test_can_set_up_a_container_that_is_already_up_and_running(self, persistent_container):
        persistent_container.up()
        assert persistent_container._container.status_code == constants.CONTAINER_RUNNING

    def test_can_set_up_a_container_that_exists_but_is_not_running(self, persistent_container):
        persistent_container.halt()
        assert persistent_container._container.status_code == constants.CONTAINER_STOPPED
        persistent_container.up()
        assert persistent_container._container.status_code == constants.CONTAINER_RUNNING

    def test_can_destroy_a_container_and_run_this_action_for_a_container_that_does_not_exist(self):
        container_options = {
            'name': self.containername('doesnotexist'), 'image': 'alpine/3.6', 'mode': 'pull', }
        container = Container('myproject', THIS_DIR, self.client, **container_options)
        container.destroy()
        assert not container.exists
        container.up()
        assert container.exists
        container.destroy()
        assert not container.exists

    def test_can_halt_a_container_that_is_running(self, persistent_container):
        persistent_container.halt()
        assert persistent_container._container.status_code == constants.CONTAINER_STOPPED

    def test_can_try_to_halt_a_container_that_is_already_stopped(self, persistent_container):
        persistent_container.halt()
        persistent_container.halt()
        assert persistent_container._container.status_code == constants.CONTAINER_STOPPED

    @pytest.mark.skipif(
        sys.version_info < (3, 5),
        reason="ansible requires python 3.5+"
    )
    @pytest.mark.parametrize("lxd_transport", [False, True])
    def test_can_provision_a_container_ansible(self, persistent_container, lxd_transport):
        # We want to make sure that barebone setup is executed on provision()
        persistent_container._container.config['user.lxdock.provisioned'] = 'false'
        persistent_container._container.save(wait=True)
        persistent_container.options['provisioning'] = [{
            'type': 'ansible',
            'playbook': os.path.join(THIS_DIR, 'fixtures/provision_with_ansible.yml'),
            'lxd_transport': lxd_transport,
        }]
        persistent_container.up()
        persistent_container.provision()
        assert persistent_container._container.files.get('/dummytest').strip() == b'dummytest'
        persistent_container._container.execute(['rm', '/dummytest'])
        del persistent_container.options['provisioning']

    def test_can_provision_a_container_shell_inline(self):
        container_options = {
            'name': self.containername('willprovision'), 'image': 'alpine/3.6', 'mode': 'pull',
            'environment': {'PATH': '/dummy_test:/bin:/usr/bin:/usr/local/bin'},
            'provisioning': [
                {'type': 'shell',
                 'inline': """touch f && echo "Here's the PATH" $PATH >> /tmp/test.txt""", }
            ],
        }
        container = Container('myproject', THIS_DIR, self.client, **container_options)
        container.up()
        container.provision()
        assert container._container.files.get('/tmp/test.txt').strip() == (
            b"Here's the PATH /dummy_test:/bin:/usr/bin:/usr/local/bin")

    @unittest.mock.patch('subprocess.call')
    def test_can_open_a_shell_for_the_root_user(self, mocked_call, persistent_container):
        persistent_container.shell()
        assert mocked_call.call_count == 1
        assert mocked_call.call_args[0][0] == \
            'lxc exec {} -- su -m root'.format(persistent_container.lxd_name)

    @unittest.mock.patch('subprocess.call')
    def test_can_open_a_shell_for_a_specific_shelluser(self, mocked_call):
        container_options = {
            'name': self.containername('shellspecificuser'), 'image': 'alpine/3.6',
            'shell': {'user': 'test', 'home': '/opt', },
        }
        container = Container('myproject', THIS_DIR, self.client, **container_options)
        container.up()
        container.shell()
        assert mocked_call.call_count == 1
        assert mocked_call.call_args[0][0] == \
            'lxc exec {} --env HOME=/opt -- su -m test'.format(container.lxd_name)

    @unittest.mock.patch('subprocess.call')
    def test_can_run_quoted_shell_command_for_the_root_user(
            self, mocked_call, persistent_container):
        persistent_container.shell(cmd_args=['echo', 'he re"s', '-u', '$PATH'])
        assert mocked_call.call_count == 1
        assert mocked_call.call_args[0][0] == \
            'lxc exec {} -- su -m root -s {}'.format(
                persistent_container.lxd_name, persistent_container._guest_shell_script_file)
        script = persistent_container._container.files.get(
            persistent_container._guest_shell_script_file)
        assert script == b"""#!/bin/sh\necho 'he re"s' -u '$PATH'\n"""

    @unittest.mock.patch('subprocess.call')
    def test_can_run_quoted_shell_command_for_a_specific_shelluser(self, mocked_call):
        container_options = {
            'name': self.containername('shellspecificuser'), 'image': 'alpine/3.6',
            'shell': {'user': 'test', 'home': '/opt', },
        }
        container = Container('myproject', THIS_DIR, self.client, **container_options)
        container.up()
        container.shell(cmd_args=['echo', 'he re"s', '-u', '$PATH'])
        assert mocked_call.call_count == 1
        assert mocked_call.call_args[0][0] == \
            'lxc exec {} --env HOME=/opt -- su -m test -s {}'.format(
                container.lxd_name, container._guest_shell_script_file)
        script = container._container.files.get(container._guest_shell_script_file)
        assert script == b"""#!/bin/sh\necho 'he re"s' -u '$PATH'\n"""

    @unittest.mock.patch('subprocess.call')
    def test_can_set_shell_environment_variables(self, mocked_call):
        # Environment variables in the shell can be set through configuration.
        container_options = {
            'name': self.containername('shell-env'), 'image': 'alpine/3.6',
            'environment': {'FOO': 'bar', 'BAR': 42},
        }
        container = Container('myproject', THIS_DIR, self.client, **container_options)
        container.up()
        container.shell()
        assert container._container.config['environment.FOO'] == 'bar'
        assert container._container.config['environment.BAR'] == '42'

    def test_can_tell_if_a_container_exists_or_not(self, persistent_container):
        unknown_container = Container('myproject', THIS_DIR, self.client, **{
            'name': self.containername('unknown'), 'image': 'alpine/3.6', 'mode': 'pull', })
        assert persistent_container.exists
        assert not unknown_container.exists

    def test_can_tell_if_a_container_is_privileged_or_not(self, persistent_container):
        persistent_container._container.config['security.privileged'] = 'true'
        persistent_container._container.save(wait=True)
        assert persistent_container.is_privileged
        persistent_container._container.config['security.privileged'] = 'false'
        persistent_container._container.save(wait=True)
        assert not persistent_container.is_privileged

    def test_can_tell_if_a_container_is_provisioned_or_not(self, persistent_container):
        persistent_container._container.config['user.lxdock.provisioned'] = 'false'
        persistent_container._container.save(wait=True)
        assert not persistent_container.is_provisioned
        persistent_container._container.config['user.lxdock.provisioned'] = 'true'
        persistent_container._container.save(wait=True)
        assert persistent_container.is_provisioned

    def test_can_tell_if_a_container_is_running_or_not(self, persistent_container):
        persistent_container.halt()
        assert not persistent_container.is_running
        persistent_container.up()
        assert persistent_container.is_running

    def test_can_tell_if_a_container_is_stopped_or_not(self, persistent_container):
        persistent_container.halt()
        assert persistent_container.is_stopped
        persistent_container.up()
        assert not persistent_container.is_stopped

    def test_can_return_its_status(self, persistent_container):
        unknown_container = Container('myproject', THIS_DIR, self.client, **{
            'name': self.containername('unknown'), 'image': 'alpine/3.6', 'mode': 'pull', })
        assert unknown_container.status == 'not-created'
        persistent_container.halt()
        assert persistent_container.status == 'stopped'
        persistent_container.up()
        assert persistent_container.status == 'running'

    def test_create_users(self):
        password = '$6$cGzZBkDjOhGW$6C9wwqQteFEY4lQ6ZJBggE568SLSS7bIMKexwOD' \
                   '39mJQrJcZ5vIKJVIfwsKOZajhbPw0.Zqd0jU2NDLAnp9J/1'
        container_options = {
            'name': self.containername('createusers'), 'image': 'alpine/3.6',
            'users': [
                {'name': 'user01'},
                {'name': 'user02', 'home': '/opt/user02'},
                {'name': 'user03', 'password': password},
            ],
        }
        container = Container('myproject', THIS_DIR, self.client, **container_options)
        guest_mock = unittest.mock.Mock()
        container._container_guest = guest_mock
        container.up()
        assert guest_mock.create_user.call_count == 3
        assert guest_mock.create_user.call_args_list[0][0][0] == 'user01'
        assert guest_mock.create_user.call_args_list[1][0][0] == 'user02'
        assert guest_mock.create_user.call_args_list[2][0][0] == 'user03'
        assert guest_mock.create_user.call_args_list[1][1]['home'] == '/opt/user02'
        assert guest_mock.create_user.call_args_list[2][1]['password'] == password

    def test_get_container_lxc_config(self):
        """Test that _get_container generates a valid lxc_config
        """

        # The options below has an lxc_config value, that overrides some values
        # that are driven within lxdoc, these values are marked as invalid and should not
        # be passed directly to the container at creation time.
        container_options = {
            'name': self.containername('lxc-config'),
            'image': 'alpine/3.6',
            'lxc_config': {
                'security.privileged': 'invalid',
                'user.lxdock.homedir': 'invalid',
                'user.lxdock.made': 'invalid',
                'valid_key': 'valid_value',
            },
        }

        cont_return = ()  # mock container object to return

        def mock_create(container_config, *args, **kwargs):
            """Mocks the container create call, returns the mock container object
            and also ensures that the container_config is correct
            """
            config = container_config['config']
            # Values below should not be driven by the values in container_options
            assert config['security.privileged'] != 'invalid'
            assert config['user.lxdock.homedir'] != 'invalid'
            assert config['user.lxdock.made'] != 'invalid'
            # Value below is a custom value that should be passed from container_options
            assert config['valid_key'] == 'valid_value'
            return cont_return

        client_mock = unittest.mock.Mock(**{
            'containers.create.side_effect': mock_create,
            'containers.get.side_effect': NotFound(''),
        })

        container = Container('myproject', THIS_DIR, client_mock, **container_options)

        assert container._get_container() is cont_return
        assert client_mock.containers.get.called
        assert client_mock.containers.create.called

    def test_can_set_profiles(self):
        container_options = {
            'name': self.containername('newcontainer'), 'image': 'alpine/3.6', 'mode': 'pull',
            'profiles': ['default']}
        container = Container('myproject', THIS_DIR, self.client, **container_options)
        container.up()
        assert container._container.status_code == constants.CONTAINER_RUNNING
        assert container._container.profiles == ['default']

    def test_raises_an_error_if_profile_does_not_exist(self):
        container_options = {
            'name': self.containername('newcontainer'), 'image': 'alpine/3.6', 'mode': 'pull',
            'profiles': ['default', '39mJQrJcZ5vIKJVIfwsKOZajhbPw0']}
        container = Container('myproject', THIS_DIR, self.client, **container_options)
        with pytest.raises(ContainerOperationFailed):
            container.up()



from .base import Guest


class DebianGuest(Guest):
    """ This guest can provision Debian containers. """

    name = 'debian'

    def install_packages(self, packages):
        self.run(['apt-get', 'update'])
        self.run(['apt-get', 'install', '-y'] + packages)



"""
    Base host
    =========
    This module provides the `Host` base class that is used to define the host OS or distributions
    supported by LXDock.
"""

import logging
import os
import platform
import shlex
import subprocess
from pathlib import Path

from ..utils.lxd import get_lxd_dir
from ..utils.metaclass import with_metaclass


__all__ = ['Host', ]

logger = logging.getLogger(__name__)


class InvalidHost(Exception):
    """ The `Host` subclass is not valid. """


class _HostBase(type):
    """ Metaclass for all LXD hosts.

    This metaclass ensures that all defined `Host` subclasses have the required attributes and
    proceeds to some validation checks. Additionally it implements the "plugin mount" paradigm and
    stores a list of `Host` subclasses in the namespace of the "plugin mount" class (`host`).
    """

    def __new__(cls, name, bases, attrs):
        super_new = super(_HostBase, cls).__new__
        parents = [base for base in bases if isinstance(base, _HostBase)]

        # We stop here if we are considering the top-level class to which this metaclass was applied
        # and not one of its subclasses (eg. Host).
        if not parents:
            return super_new(cls, name, bases, attrs)

        # Constructs the Host class.
        new_host = super_new(cls, name, bases, attrs)

        # Performs some validation checks.
        if not new_host.name:
            raise InvalidHost("The 'name' attribute of Host subclasses cannot be None")

        return new_host

    def __init__(cls, name, bases, attrs):
        # We implement the "mount point" paradigm here in order to make all `Host` subclassses
        # available in a single attribute.
        if not hasattr(cls, 'hosts'):
            # The class has no hosts list: this means that we are considering the "plugin mount"
            # class. So we created the list that will hold all the defined `Host` subclasses.
            cls.hosts = []
        else:
            # The `hosts` attribute already exists so we are considering a `Host` subclass, which
            # needs to be registered.
            cls.hosts.append(cls)


class Host(with_metaclass(_HostBase)):
    """ Represents a single host.

    `Host` subclasses will be used by `Container` instances to perform common operations on the
    host side. For example they can be used to retrieve some date (SSH pukeys, ...) or to set up
    contaianers' hosts in the /etc/hosts file. `Host` subclasses should correspond to specific OSes
    or distributions that can be used to run LXD and LXDock.
    """

    # The `name` of a host is a required attribute and should always be set on `Host` subclasses.
    name = None

    @classmethod
    def detect(cls):
        """ Detects if the host is an "instance" of the considered OS/distribution. """
        return cls.name.lower() in platform.platform().lower()

    @classmethod
    def get(cls):
        """ Returns the `Host` instance associated with the considered host. """
        class_ = next((k for k in Host.hosts if k.detect()), Host)
        return class_()

    def get_ssh_pubkey(self):
        """ Returns the SSH public key of the current user or None if it cannot be found. """
        pubkey_path = Path(os.path.expanduser('~/.ssh/id_rsa.pub'))
        try:
            return pubkey_path.open().read()
        except FileNotFoundError:  # pragma: no cover
            pass

    def give_current_user_access_to_share(self, source):
        """ Give read/write access to `source` for the current user. """
        self.run(['setfacl', '-Rdm',  'u:{}:rwX'.format(os.getuid()), source])

    def give_mapped_user_access_to_share(self, lxd_container, source, userpath=None):
        """ Give read/write access to `source` for the mapped user owning `userpath`.

        `userpath` is a path that is relative to the LXD base directory (where LXD store contaners).
        """
        # LXD uses user namespaces when running safe containers. This means that it maps a set of
        # uids and gids on the host to a set of uids and gids in the container.
        # When considering unprivileged containers we want to ensure that the "root user" (or any
        # other user) of such containers have the proper rights to write in shared folders. To do so
        # we have to retrieve the UserID on the host-side that is mapped to the "root"'s UserID (or
        # any other user's UserID) on the guest-side. This will allow to set ACL on the host-side
        # for this UID. By doing this we will also allow "root" user on the guest-side to read/write
        # in shared folders.
        container_path_parts = [get_lxd_dir(), 'containers', lxd_container.name, 'rootfs']
        container_path_parts += userpath.split('/') if userpath else []
        container_path = os.path.join(*container_path_parts)
        container_path_stats = os.stat(container_path)
        host_userpath_uid = container_path_stats.st_uid
        self.run([
            'setfacl', '-Rm',
            'user:lxd:rwx,default:user:lxd:rwx,'
            'user:{0}:rwx,default:user:{0}:rwx'.format(host_userpath_uid), source,
        ])

    ##################
    # HELPER METHODS #
    ##################

    def run(self, cmd_args):
        """ Runs the specified command on the host. """
        cmd = ' '.join(map(shlex.quote, cmd_args))
        logger.debug('Running {0} on the host'.format(cmd))
        subprocess.Popen(cmd, shell=True).wait()



"""
    LXDock specific pytest fixtures
    ===============================
    This module provides fixtures that can be used to perform common operations before or after test
    execution.
"""

import os

import pytest

from ..client import get_client
from ..container import Container


__all__ = [
    'persistent_container', 'remove_persistent_container',
]


THIS_DIR = os.path.join(os.path.dirname(__file__))

_persistent_container = None


@pytest.fixture
def persistent_container():
    """ Returns a persistent `lxdock.container.Container` instance.

    This container will be removed when tearing down the test class. Of course this container should
    not be removed by test methods (in that case it will be recreated) in order to speed up test
    execution.
    """
    global _persistent_container
    if _persistent_container is None:
        _persistent_container = Container(
            'lxdtestcase-persistentcontainer', THIS_DIR, get_client(), **{
                'name': 'testcase-persistent', 'image': 'alpine/3.6', 'mode': 'pull',
            })
    # Ensures the persistent container is up and running.
    if not _persistent_container.exists \
            or (_persistent_container.exists and _persistent_container.is_stopped):
        _persistent_container.up()
    return _persistent_container


@pytest.fixture(scope='session', autouse=True)
def remove_persistent_container():
    """ Removes the directories inside the MEDIA_ROOT that could have been filled during tests. """
    yield
    # Removes the persistent container after tests execution.
    global _persistent_container
    if _persistent_container is not None:
        _persistent_container.destroy()
    del _persistent_container



import re
import unittest.mock

import pytest

from lxdock.guests import AlpineGuest, DebianGuest
from lxdock.hosts import Host
from lxdock.provisioners import AnsibleProvisioner
from lxdock.test import FakeContainer


class TestAnsibleProvisioner:
    @pytest.mark.parametrize("options,expected_cmdargs", [
        ({}, ''),
        ({'vault_password_file': '.vpass'}, '--vault-password-file ./.vpass'),
        ({'ask_vault_pass': True}, '--ask-vault-pass'),
        ({'lxd_transport': True}, '-c lxd'),
    ])
    @unittest.mock.patch('subprocess.Popen')
    def test_can_run_ansible_playbooks(self, mock_popen, options, expected_cmdargs):
        host = Host()
        guest = DebianGuest(FakeContainer())
        lxd_state = unittest.mock.Mock()
        lxd_state.network.__getitem__ = unittest.mock.MagicMock(
            return_value={'addresses': [{'family': 'init', 'address': '0.0.0.0', }, ]})
        guest.lxd_container.state.return_value = lxd_state
        options['playbook'] = 'deploy.yml'
        provisioner = AnsibleProvisioner('./', host, [guest], options)
        provisioner.provision()
        m = re.match(
            r'ANSIBLE_HOST_KEY_CHECKING=False ansible-playbook --inventory-file /[/\w]+ '
            '(.*)\s?./deploy.yml', mock_popen.call_args[0][0])
        assert m
        assert m.group(1).strip() == expected_cmdargs

    def test_can_properly_setup_ssh_for_alpine_guests(self):
        container = FakeContainer()
        lxd_container = container._container
        lxd_container.execute.return_value = ('ok', 'ok', '')
        host = Host()
        guest = AlpineGuest(container)
        provisioner = AnsibleProvisioner('./', host, [guest], {'playbook': 'deploy.yml'})
        provisioner.setup()
        EXPECTED = [
            ['apk', 'update'],
            ['apk', 'add'] + AnsibleProvisioner.guest_required_packages_alpine,
            ['apk', 'update'],
            ['apk', 'add', 'openssh'],
            ['rc-update', 'add', 'sshd'],
            ['/etc/init.d/sshd', 'start'],
            ['mkdir', '-p', '/root/.ssh'],
        ]
        calls = [tup[0][0] for tup in lxd_container.execute.call_args_list]
        assert calls == EXPECTED

    def test_inventory_contains_groups(self):
        c1 = FakeContainer(name='c1')
        c2 = FakeContainer(name='c2')
        # c3 is deliberately not part of our guests list. This is to test that it doesn't end up
        # in the inventory and result in spurious unreachable hosts. These situations can happen
        # in two ways: errors in the config file, or guest filtering in the command line ("lxdock
        # provision c1, c2" for example).
        provisioner = AnsibleProvisioner(
            './',
            Host(),
            [DebianGuest(c1), DebianGuest(c2)],
            {'playbook': 'deploy.yml', 'groups': {'g1': ['c1', 'c2'], 'g2': ['c1', 'c3']}}
        )
        inv = provisioner.get_inventory()
        # group order is not guaranteed. our tests have to be written with that in mind.
        groups = re.findall(r'\[(g1|g2)\]([^[]+)', inv, re.MULTILINE)
        # we sort so that g1 will be first all the time
        groups = sorted([(gname, hosts.strip()) for gname, hosts in groups])
        assert sorted(groups) == [('g1', 'c1\nc2'), ('g2', 'c1')]

    def test_inventory_with_lxd_transport(self):
        c = FakeContainer(name='c1')
        provisioner = AnsibleProvisioner(
            './',
            Host(),
            [DebianGuest(c)],
            {'playbook': 'deploy.yml', 'lxd_transport': True}
        )
        inv = provisioner.get_inventory()
        assert 'ansible_host={}'.format(c.lxd_name) in inv



import os
import tempfile
import unittest.mock
from pathlib import Path, PurePosixPath

import pytest
from voluptuous import Schema
from voluptuous.error import Invalid

from lxdock.exceptions import ProvisionFailed
from lxdock.guests import Guest
from lxdock.hosts import Host
from lxdock.provisioners import PuppetProvisioner


class TestPuppetProvisionerSchema:
    def test_can_set_default_manifest_file(self):
        with tempfile.TemporaryDirectory() as d:
            with (Path(d) / 'default.pp').open('w') as f:
                f.write('dummy pp file')
            config = {'manifests_path': d}
            try:
                final_config = Schema(PuppetProvisioner.schema)(config)
            except Invalid as e:
                pytest.fail("schema validation didn't pass: {}".format(e))
            assert final_config == {
                'manifest_file': 'default.pp',
                'manifests_path': d}

    def test_can_set_default_environment(self):
        with tempfile.TemporaryDirectory() as d:
            os.mkdir(str(Path(d) / 'production'))
            config = {'environment_path': d}
            try:
                final_config = Schema(PuppetProvisioner.schema)(config)
            except Invalid as e:
                pytest.fail("schema validation didn't pass: {}".format(e))
            assert final_config == {
                'environment': 'production',
                'environment_path': d}

    @unittest.mock.patch('lxdock.provisioners.puppet.IsDir')
    @unittest.mock.patch('lxdock.provisioners.puppet.IsFile')
    def test_can_set_default_mode_and_manifest_file(self, mock_isfile, mock_isdir):
        config = {}
        try:
            final_config = Schema(PuppetProvisioner.schema)(config)
        except Invalid as e:
            pytest.fail("schema validation didn't pass: {}".format(e))
        assert final_config == {
            'manifests_path': 'manifests',
            'manifest_file': 'default.pp'}


class TestPuppetProvisioner:
    @unittest.mock.patch.object(Guest, 'copy_directory')
    @unittest.mock.patch.object(Guest, 'run')
    def test_can_run_puppet_manifest_mode(self, mock_run, mock_copy_dir):
        class DummyGuest(Guest):
            name = 'dummy'
        host = Host()
        guest = DummyGuest(unittest.mock.Mock())
        mock_run.return_value = 0
        provisioner = PuppetProvisioner('./', host, [guest], {
            'manifest_file': 'test_site.pp',
            'manifests_path': 'test_manifests'})
        provisioner.provision()

        assert mock_copy_dir.call_count == 1
        assert mock_copy_dir.call_args_list[0][0][0] == Path('test_manifests')
        assert mock_copy_dir.call_args_list[0][0][1] == PurePosixPath(
            provisioner._guest_manifests_path)

        assert mock_run.call_count == 2
        assert mock_run.call_args_list[0][0][0] == ['which', 'puppet']
        assert mock_run.call_args_list[1][0][0] == [
            'sh', '-c',
            'puppet apply --detailed-exitcodes --manifestdir {} {}'.format(
                PurePosixPath(provisioner._guest_manifests_path),
                PurePosixPath(provisioner._guest_manifests_path) / 'test_site.pp')]

    @unittest.mock.patch.object(Guest, 'copy_directory')
    @unittest.mock.patch.object(Guest, 'run')
    def test_can_run_puppet_environment_mode(self, mock_run, mock_copy_dir):
        class DummyGuest(Guest):
            name = 'dummy'
        host = Host()
        guest = DummyGuest(unittest.mock.Mock())
        mock_run.return_value = 0
        provisioner = PuppetProvisioner('./', host, [guest], {
            'environment': 'test_production',
            'environment_path': 'test_environments'})
        provisioner.provision()

        assert mock_copy_dir.call_count == 1
        assert mock_copy_dir.call_args_list[0][0][0] == Path('test_environments')
        assert mock_copy_dir.call_args_list[0][0][1] == PurePosixPath(
            provisioner._guest_environment_path)

        assert mock_run.call_count == 2
        assert mock_run.call_args_list[0][0][0] == ['which', 'puppet']
        assert mock_run.call_args_list[1][0][0] == [
            'sh', '-c',
            'puppet apply --detailed-exitcodes '
            '--environmentpath {} --environment {}'.format(
                PurePosixPath(provisioner._guest_environment_path),
                'test_production')]

    @unittest.mock.patch.object(Guest, 'copy_file')
    @unittest.mock.patch.object(Guest, 'run')
    def test_raise_error_if_puppet_is_not_found(self, mock_run, mock_copy_file):
        class DummyGuest(Guest):
            name = 'dummy'
        host = Host()
        guest = DummyGuest(unittest.mock.Mock())
        mock_run.return_value = 1  # Mock the error
        provisioner = PuppetProvisioner('./', host, [guest], {
            'manifest_file': 'test_site.pp',
            'manifests_path': 'test_manifests'})
        with pytest.raises(ProvisionFailed):
            provisioner.provision()
        assert mock_run.call_count == 1
        assert mock_run.call_args[0] == (['which', 'puppet'], )
        assert mock_copy_file.call_count == 0

    @unittest.mock.patch.object(Guest, 'copy_directory')
    @unittest.mock.patch.object(Guest, 'run')
    def test_can_set_binary_path(self, mock_run, mock_copy_dir):
        class DummyGuest(Guest):
            name = 'dummy'
        host = Host()
        guest = DummyGuest(unittest.mock.Mock())
        mock_run.return_value = 0
        provisioner = PuppetProvisioner('./', host, [guest], {
            'manifest_file': 'test_site.pp',
            'manifests_path': 'test_manifests',
            'binary_path': '/test/path'})
        provisioner.provision()

        assert mock_run.call_count == 2
        assert mock_run.call_args_list[0][0][0] == ['test', '-x', '/test/path/puppet']
        assert mock_run.call_args_list[1][0][0] == [
            'sh', '-c',
            '/test/path/puppet apply --detailed-exitcodes --manifestdir {} {}'.format(
                PurePosixPath(provisioner._guest_manifests_path),
                PurePosixPath(provisioner._guest_manifests_path) / 'test_site.pp')]

    @unittest.mock.patch.object(Guest, 'copy_directory')
    @unittest.mock.patch.object(Guest, 'run')
    def test_can_set_facter(self, mock_run, mock_copy_dir):
        class DummyGuest(Guest):
            name = 'dummy'
        host = Host()
        guest = DummyGuest(unittest.mock.Mock())
        mock_run.return_value = 0
        provisioner = PuppetProvisioner('./', host, [guest], {
            'manifest_file': 'site.pp',
            'manifests_path': 'test_manifests',
            'facter': {'foo': 'bah', 'bar': 'baz baz'}})
        provisioner.provision()
        assert mock_run.call_count == 2
        assert mock_run.call_args_list[1][0][0] == [
            'sh', '-c',
            "FACTER_bar='baz baz' FACTER_foo=bah "
            "puppet apply --detailed-exitcodes --manifestdir {} {}".format(
                PurePosixPath(provisioner._guest_manifests_path),
                PurePosixPath(provisioner._guest_manifests_path) / 'site.pp')]

    @unittest.mock.patch.object(Guest, 'copy_file')
    @unittest.mock.patch.object(Guest, 'copy_directory')
    @unittest.mock.patch.object(Guest, 'run')
    def test_can_set_hiera_config_path(self, mock_run, mock_copy_dir, mock_copy_file):
        class DummyGuest(Guest):
            name = 'dummy'
        host = Host()
        guest = DummyGuest(unittest.mock.Mock())
        mock_run.return_value = 0
        provisioner = PuppetProvisioner('./', host, [guest], {
            'manifest_file': 'site.pp',
            'manifests_path': 'test_manifests',
            'hiera_config_path': 'hiera.yaml'})
        provisioner.provision()

        assert mock_copy_file.call_count == 1
        assert mock_copy_file.call_args_list[0][0][0] == Path('hiera.yaml')
        assert mock_copy_file.call_args_list[0][0][1] == PurePosixPath(
            provisioner._guest_hiera_file)

        assert mock_run.call_count == 2
        assert mock_run.call_args_list[1][0][0] == [
            'sh', '-c',
            "puppet apply --hiera_config={} --detailed-exitcodes --manifestdir {} {}".format(
                PurePosixPath(provisioner._guest_hiera_file),
                PurePosixPath(provisioner._guest_manifests_path),
                PurePosixPath(provisioner._guest_manifests_path) / 'site.pp')]

    @unittest.mock.patch.object(Guest, 'copy_directory')
    @unittest.mock.patch.object(Guest, 'run')
    def test_can_set_module_path(self, mock_run, mock_copy_dir):
        class DummyGuest(Guest):
            name = 'dummy'
        host = Host()
        guest = DummyGuest(unittest.mock.Mock())
        mock_run.return_value = 0
        provisioner = PuppetProvisioner('./', host, [guest], {
            'manifest_file': 'mani.pp',
            'manifests_path': 'test_manifests',
            'module_path': 'test-puppet-modules'})
        provisioner.provision()

        assert mock_copy_dir.call_count == 2
        assert (Path('test-puppet-modules'),
                PurePosixPath(provisioner._guest_module_path)) in {
                    mock_copy_dir.call_args_list[0][0],
                    mock_copy_dir.call_args_list[1][0]}

        assert mock_run.call_count == 2
        assert mock_run.call_args_list[1][0][0] == [
            'sh', '-c',
            "puppet apply --modulepath {}:{} --detailed-exitcodes --manifestdir {} {}".format(
                PurePosixPath(provisioner._guest_module_path),
                PurePosixPath(provisioner._guest_default_module_path),
                PurePosixPath(provisioner._guest_manifests_path),
                PurePosixPath(provisioner._guest_manifests_path) / 'mani.pp')]

    @unittest.mock.patch.object(Guest, 'copy_directory')
    @unittest.mock.patch.object(Guest, 'run')
    def test_can_set_environment_variables(self, mock_run, mock_copy_dir):
        class DummyGuest(Guest):
            name = 'dummy'
        host = Host()
        guest = DummyGuest(unittest.mock.Mock())
        mock_run.return_value = 0
        provisioner = PuppetProvisioner('./', host, [guest], {
            'manifest_file': 'site.pp',
            'manifests_path': 'test_manifests',
            'environment_variables': {'FOO': 'bah', 'BAR': 'baz baz'}})
        provisioner.provision()
        assert mock_run.call_count == 2
        assert mock_run.call_args_list[1][0][0] == [
            'sh', '-c',
            "BAR='baz baz' FOO=bah "
            "puppet apply --detailed-exitcodes --manifestdir {} {}".format(
                PurePosixPath(provisioner._guest_manifests_path),
                PurePosixPath(provisioner._guest_manifests_path) / 'site.pp')]

    @unittest.mock.patch.object(Guest, 'copy_directory')
    @unittest.mock.patch.object(Guest, 'run')
    def test_can_set_options(self, mock_run, mock_copy_dir):
        class DummyGuest(Guest):
            name = 'dummy'
        host = Host()
        guest = DummyGuest(unittest.mock.Mock())
        mock_run.return_value = 0
        provisioner = PuppetProvisioner('./', host, [guest], {
            'manifest_file': 'site.pp',
            'manifests_path': 'test_manifests',
            'options': '--a --c="test space" --b'})
        provisioner.provision()
        assert mock_run.call_count == 2
        assert mock_run.call_args_list[1][0][0] == [
            'sh', '-c',
            """puppet apply --a '--c=test space' --b """
            """--detailed-exitcodes --manifestdir {} {}""".format(
                PurePosixPath(provisioner._guest_manifests_path),
                PurePosixPath(provisioner._guest_manifests_path) / 'site.pp')]



import argparse
import logging
import sys

from .. import __version__
from ..conf.exceptions import ConfigError
from ..constants import ProvisioningMode
from ..exceptions import LXDockException
from ..logging import console_stderr_handler, console_stdout_handler
from .exceptions import CLIError


logger = logging.getLogger(__name__)


class LXDock:
    """ Wrapper around the LXDock argument parser and the related actions. """

    def __init__(self, argv=None):
        self._parsers = {}

        # Creates the argument parsers
        parser = argparse.ArgumentParser(
            description='Orchestrate and run multiple containers using LXD.', prog='lxdock')
        parser.add_argument(
            '--version', action='version', version='%(prog)s {v}'.format(v=__version__))
        parser.add_argument('-v', '--verbose', action='store_true')
        self._parsers['main'] = parser

        subparsers = parser.add_subparsers(dest='action')

        # Creates the 'config' action.
        self._parsers['config'] = subparsers.add_parser(
            'config', help='Validate and show the LXDock file.',
            description='Validate and show the LXDock file of the current project.')
        self._parsers['config'].add_argument(
            '--containers', action='store_true', help='Display only container names, one per line.')

        # Creates the 'destroy' action.
        self._parsers['destroy'] = subparsers.add_parser(
            'destroy', help='Stop and remove containers.',
            description='Destroy all the containers of a project or destroy specific containers '
                        'if container names are specified.')
        self._parsers['destroy'].add_argument(
            '-f', '--force', action='store_true', help='Destroy without confirmation.')

        # Creates the 'halt' action.
        self._parsers['halt'] = subparsers.add_parser(
            'halt', help='Stop containers.',
            description='Stop all the containers of a project or stop specific containers if '
                        'container names are specified.')

        # Creates the 'help' action.
        self._parsers['help'] = subparsers.add_parser(
            'help', help='Show help information.',
            description='Show general help information or show help information about a specific '
                        'subcommand.')
        self._parsers['help'].add_argument('subcommand', nargs='?', help='Subcommand name.')

        # Creates the 'init' action.
        self._parsers['init'] = subparsers.add_parser(
            'init', help='Generate a LXDock file.',
            description='Generate a LXDock file defining a single container and highlighting '
                        'useful options.')
        self._parsers['init'].add_argument(
            '-f', '--force', action='store_true', help='Overwrite existing LXDock file')
        self._parsers['init'].add_argument('--image', help='Container image to use')
        self._parsers['init'].add_argument('--project', help='Project name to use')

        # Creates the 'provision' action.
        self._parsers['provision'] = subparsers.add_parser(
            'provision', help='Provision containers.',
            description='Provision all the containers of a project or provision specific '
                        'containers if container names are specified.')

        # Creates the 'shell' action.
        self._parsers['shell'] = subparsers.add_parser(
            'shell', help='Open a shell or execute a command in a container.',
            description='Open an interactive shell inside a specific container. If a command is '
                        'specified, execute the command instead.',
            usage='lxdock shell [-h] [-u USERNAME] [name] [-c ...]')
        self._parsers['shell'].add_argument('name', nargs='?', help='Container name.')
        self._parsers['shell'].add_argument(
            '-u', '--username', help='Username to login as.')
        self._parsers['shell'].add_argument(
            '-c', '--command', nargs=argparse.REMAINDER, dest='cmd_args',
            help='Command to be executed.')

        # Creates the 'status' action.
        self._parsers['status'] = subparsers.add_parser(
            'status', help='Show containers\' statuses.',
            description='Show the status of all the containers of a project or show the status of '
                        'specific containers if container names are specified.')

        # Creates the 'up' action.
        self._parsers['up'] = subparsers.add_parser(
            'up', help='Create, start and provision containers.',
            description='Create, start and provision all the containers of the project according '
                        'to your LXDock file. If container names are specified, only the related '
                        'containers are created, started and provisioned.')
        up_provision_group = self._parsers['up'].add_mutually_exclusive_group()
        up_provision_group.add_argument(
            '--provision', action='store_const', const=ProvisioningMode.ENABLED,
            dest='provisioning_mode', help='Enable and force provisioning.')
        up_provision_group.add_argument(
            '--no-provision', action='store_const', const=ProvisioningMode.DISABLED,
            dest='provisioning_mode', help='Disable provisioning.')
        up_provision_group.set_defaults(provisioning_mode=None)

        # Add common arguments to the action parsers that can be used with one or more specific
        # containers.
        per_container_parsers = ['destroy', 'halt', 'provision', 'status', 'up', ]
        for pkey in per_container_parsers:
            self._parsers[pkey].add_argument('name', nargs='*', help='Container name.')

        # Parses the arguments
        args = parser.parse_args(args=argv)

        # Displays the help if no action is specified
        if args.action is None:
            parser.print_help()
            sys.exit(1)

        # Handles verbosity
        if args.verbose:
            console_stdout_handler.setLevel(logging.DEBUG)
        else:
            console_stdout_handler.setLevel(logging.INFO)

        try:
            # use dispatch pattern to invoke method with same name
            getattr(self, args.action)(args)
        except KeyboardInterrupt:
            logger.warn('\nAborting.')
            sys.exit(1)
        except (CLIError, ConfigError, LXDockException) as e:
            if e.msg is not None:
                logger.error(e.msg)
                sys.exit(1)

    def config(self, args):
        # We have to display the LXDock file here, which can be useful for completion or other
        # automated operations. In order to speed up things, we'll just manually create our config
        # object and use it.
        if args.containers:
            # This option is mostly used for completion features. Thus we use print directly in
            # order to not embed control characters in the final output.
            print('\n'.join([c['name'] for c in self.project_config.containers]))
            return

        logger.info(self.project_config.serialize())

    def destroy(self, args):
        from .utils import yesno
        # Builds a list of containers by ensuring that these containers are actually associated with
        # the considered project.
        container_names = args.name or [c.name for c in self.project.containers]
        containers = [self.project.get_container_by_name(name) for name in container_names]
        # At this point we are sure that the containers we are manipulating are defined for the
        # project because we used the `get_container_by_name` method, which raises an error if a
        # container is not defined for the considered project.
        existing_containers = [c for c in containers if c.exists]

        should_destroy = False
        if len(container_names) and not len(existing_containers):
            should_destroy = True
        else:
            # We display a confirmation input to the user only if there are existing containers for
            # the considered project. Otherwise we just go through the destroy methods of the
            # containers and let the logger tell the user that the containers were not created.
            logging.warning(
                'The following containers will be removed: {}'.format(', '.join(container_names)))
            if len(container_names) and (args.force or yesno('Are you sure?')):
                should_destroy = True

        if should_destroy:
            self.project.destroy(container_names=args.name)

    def halt(self, args):
        self.project.halt(container_names=args.name)

    def help(self, args):
        try:
            assert args.subcommand is not None
            self._parsers[args.subcommand].print_help()
        except AssertionError:
            self._parsers['main'].print_help()
        except KeyError:
            # args.subcommand is not a valid subcommand!
            raise CLIError('No such command: {}'.format(args.subcommand))

    def init(self, args):
        import os
        from ..conf.constants import ALLOWED_FILENAMES
        from .constants import INIT_LXDOCK_FILE_CONTENT
        cwd = os.getcwd()
        project_name = args.project or os.path.split(cwd)[1]

        # Check if an existing config file is already present in the current working directory.
        existing_config = [
            filename for filename in ALLOWED_FILENAMES
            if os.path.exists(os.path.join(cwd, filename))]
        if existing_config and not args.force:
            raise CLIError(
                'An existing LXDock file is already present in this directory! Using "lxdock init" '
                'could overwrite this file. Use the -f/--force to overwrite existing LXDock files.')

        # Compute the content of the LXDock file to write and write it to a lxdock.yml file.
        init_filecontent = INIT_LXDOCK_FILE_CONTENT.format(
            project_name=project_name, image=args.image or 'ubuntu/xenial')
        with open('lxdock.yml', mode='w', encoding='utf-8') as fd:
            fd.write(init_filecontent)

    def provision(self, args):
        self.project.provision(container_names=args.name)

    def shell(self, args):
        self.project.shell(
            container_name=args.name, username=args.username, cmd_args=args.cmd_args)

    def status(self, args):
        self.project.status(container_names=args.name)

    def up(self, args):
        self.project.up(container_names=args.name, provisioning_mode=args.provisioning_mode)

    ##################################
    # UTILITY METHODS AND PROPERTIES #
    ##################################

    @property
    def project(self):
        """ Initializes a LXDock project instance and returns it. """
        from .project import get_project
        if not hasattr(self, '_project'):
            self._project = get_project()
        return self._project

    @property
    def project_config(self):
        """ Initializes a `Config` instance and returns it. """
        from ..conf import Config
        if not hasattr(self, '_project_config'):
            self._project_config = Config.from_base_dir()
        return self._project_config


def main(argv=None):
    # Setup logging
    root_logger = logging.getLogger()
    root_logger.addHandler(console_stdout_handler)
    root_logger.addHandler(console_stderr_handler)
    root_logger.setLevel(logging.DEBUG)
    # Disables requests logging
    logging.getLogger('requests').propagate = False
    logging.getLogger('ws4py').propagate = False

    # Run the LXDock orchestration tool!
    LXDock(argv=argv)



"""
    The ``constants`` module
    ========================

    This module defines top-level constants that can be used all tools or classes provided by
    LXDock, such as projects, containers or CLI parsers.

"""

from enum import Enum


# CONTAINER STATUSES
# --

CONTAINER_STOPPED = 102
CONTAINER_RUNNING = 103


# PROVISIONING
# --

class ProvisioningMode(Enum):
    """ Defines some common values that can be used to identify provisioning behavior types. """

    # In "auto" mode containers will be provisioned only at creation time.
    AUTO = 1

    # In "enabled" mode containers will be systematically provisioned.
    ENABLED = 2

    # In "disabled" mode containers won't be provisioned.
    DISABLED = 3



from .base import Guest


class GentooGuest(Guest):
    """ This guest can provision Gentoo containers. """

    name = 'gentoo'

    def install_packages(self, packages):
        # Ensure that we have this Gentoo helper toolkit.
        # It contains "equery" that can check which package has been installed.
        self.run(['emerge', 'app-portage/gentoolkit'])
        for p in packages:
            retcode = self.run(['equery', 'list', p])
            if retcode != 0:  # Not installed yet
                self.run(['emerge', p])



import logging
import shlex
from pathlib import Path, PurePosixPath

from voluptuous import All, IsDir, IsFile

from ..exceptions import ProvisionFailed
from .base import Provisioner


__all__ = ('PuppetProvisioner', )

logger = logging.getLogger(__name__)


def finalize_options(options):
    # Refs Vagrant code:
    # https://github.com/mitchellh/vagrant/blob/9c299a2a357fcf87f356bb9d56e18a037a53d138/
    #         plugins/provisioners/puppet/config/puppet.rb#L58
    if options.get('environment_path') is None and options.get('manifests_path') is None:
        logger.warn("environment_path and manifests_path are both unset, "
                    "assuming manifests mode and manifests_path as 'manifests'...")
        options['manifests_path'] = 'manifests'

    if options.get('environment_path') is None:
        if options.get('manifest_file') is None:
            logger.warn("manifest_file is not set. Assuming 'default.pp'...")
            options['manifest_file'] = 'default.pp'
    else:
        if options.get('environment') is None:
            logger.warn("environment is not set. Assuming 'production'...")
            options['environment'] = 'production'

    return options


def validate_paths(options):
    # Refs Vagrant code:
    # https://github.com/mitchellh/vagrant/blob/9c299a2a357fcf87f356bb9d56e18a037a53d138/
    #         plugins/provisioners/puppet/config/puppet.rb#L112
    if options.get('manifests_path') is not None:
        host_manifest_file = str(
            Path(options['manifests_path']) / options['manifest_file'])
        IsFile(msg="File {} does not exist".format(host_manifest_file))(host_manifest_file)
    elif options.get('environment_path') is not None:
        host_selected_environment_path = str(
            Path(options['environment_path']) / options['environment'])
        IsDir(msg="Directory {} does not exist".format(host_selected_environment_path))(
            host_selected_environment_path)
    return options


class PuppetProvisioner(Provisioner):
    """ Allows to perform provisioning operations using Puppet. """

    name = 'puppet'

    guest_required_packages_arch = ['puppet']
    guest_required_packages_debian = ['puppet']
    guest_required_packages_fedora = ['puppet']
    guest_required_packages_ubuntu = ['puppet']

    # Refs Vagrant docs:
    # https://www.vagrantup.com/docs/provisioning/puppet_apply.html#options
    schema = All({
        'binary_path': str,
        'facter': dict,
        'hiera_config_path': IsFile(),
        'manifest_file': str,
        'manifests_path': IsDir(),
        'module_path': IsDir(),
        'environment': str,
        'environment_path': IsDir(),
        'environment_variables': dict,
        'options': str,
    }, finalize_options, validate_paths)

    _guest_manifests_path = '/.lxdock.d/puppet/manifests'
    _guest_module_path = '/.lxdock.d/puppet/modules'
    _guest_default_module_path = '/etc/puppet/modules'
    _guest_environment_path = '/.lxdock.d/puppet/environments'
    _guest_hiera_file = '/.lxdock.d/puppet/hiera.yaml'

    def provision_single(self, guest):
        """ Performs the provisioning operations using puppet. """
        # Verify if `puppet` has been installed.
        binary_path = self.options.get('binary_path')
        if binary_path is not None:
            puppet_bin = str(PurePosixPath(binary_path) / 'puppet')
            retcode = guest.run(['test', '-x', puppet_bin])
            fail_msg = (
                "puppet executable is not found in the specified path {} in the "
                "guest container. ".format(binary_path)
            )
        else:
            retcode = guest.run(['which', 'puppet'])
            fail_msg = (
                "puppet was not automatically installed for this guest. "
                "Please specify the command to install it in LXDock file using "
                "a shell provisioner and try `lxdock provision` again. You may "
                "also specify `binary_path` that contains the puppet executable "
                "in LXDock file.")
        if retcode != 0:
            raise ProvisionFailed(fail_msg)

        # Copy manifests dir
        manifests_path = self.options.get('manifests_path')
        if manifests_path is not None:
            guest.copy_directory(
                Path(manifests_path), PurePosixPath(self._guest_manifests_path))

        # Copy module dir
        module_path = self.options.get('module_path')
        if module_path is not None:
            guest.copy_directory(Path(module_path), PurePosixPath(self._guest_module_path))

        # Copy environment dir
        environment_path = self.options.get('environment_path')
        if environment_path is not None:
            guest.copy_directory(
                Path(environment_path), PurePosixPath(self._guest_environment_path))

        # Copy hiera file
        hiera_file = self.options.get('hiera_config_path')
        if hiera_file is not None:
            guest.copy_file(Path(hiera_file), PurePosixPath(self._guest_hiera_file))

        # Run puppet.
        command = self._build_puppet_command()

        if environment_path:
            logger.info("Running Puppet with environment {}...".format(self.options['environment']))
        else:
            logger.info("Running Puppet with {}...".format(self.options['manifest_file']))

        guest.run(['sh', '-c', ' '.join(command)])

    ##################################
    # PRIVATE METHODS AND PROPERTIES #
    ##################################

    def _build_puppet_command(self):
        """
        Refs:
        https://github.com/mitchellh/vagrant/blob/9c299a2a357fcf87f356bb9d56e18a037a53d138/
                plugins/provisioners/puppet/provisioner/puppet.rb#L173
        """

        options = self.options.get('options', '')
        options = list(map(shlex.quote, shlex.split(options)))

        module_path = self.options.get('module_path')
        if module_path is not None:
            options += ['--modulepath',
                        '{}:{}'.format(self._guest_module_path,
                                       self._guest_default_module_path)]

        hiera_path = self.options.get('hiera_config_path')
        if hiera_path is not None:
            options += ['--hiera_config={}'.format(self._guest_hiera_file)]

        # TODO: we are not detecting console color support now, but please contribute if you need!

        options += ['--detailed-exitcodes']

        environment_path = self.options.get('environment_path')
        if environment_path is not None:
            options += ['--environmentpath', str(self._guest_environment_path),
                        '--environment', self.options['environment']]
        else:
            options += ['--manifestdir', str(self._guest_manifests_path)]

        manifest_file = self.options.get('manifest_file')
        if manifest_file is not None:
            options += [str(
                PurePosixPath(self._guest_manifests_path) / manifest_file)]

        # Build up the custom facts if we have any
        facter = []
        facter_dict = self.options.get('facter')
        if facter_dict is not None:
            for key, value in sorted(facter_dict.items()):
                facter.append("FACTER_{}={}".format(key, shlex.quote(value)))

        binary_path = self.options.get('binary_path')
        if binary_path is not None:
            puppet_bin = str(PurePosixPath(binary_path) / 'puppet')
        else:
            puppet_bin = 'puppet'

        # TODO: working_directory for hiera. Please contribute!

        env = []
        env_variables = self.options.get('environment_variables')
        if env_variables is not None:
            for key, value in sorted(env_variables.items()):
                env.append("{}={}".format(key, shlex.quote(value)))

        command = env + facter + [puppet_bin, 'apply'] + options

        return command



class ConfigError(Exception):
    """ An error related to the configuration of the LXDock project occured. """

    def __init__(self, msg):
        self.msg = msg


class ConfigFileNotFoundError(ConfigError):
    """ The LXDock config file was not found. """


class ConfigFileValidationError(ConfigError):
    """ The LXDock config file was not valid. """


class ConfigFileInterpolationError(ConfigError):
    """ The variable substitution mechanism failed for the considered LXDock file. """



from voluptuous import (ALLOW_EXTRA, All, Any, Coerce, Extra, In, IsDir, Length, Required, Schema,
                        Url)

from ..provisioners import Provisioner
from .validators import Hostname, LXDIdentifier


def get_schema():
    _top_level_and_containers_common_options = {
        'environment': {Extra: Coerce(str)},
        'hostnames': [Hostname(), ],
        'image': str,
        'lxc_config': {Extra: str},
        'mode': In(['local', 'pull', ]),
        'privileged': bool,
        'profiles': [str, ],
        'protocol': In(['lxd', 'simplestreams', ]),
        'provisioning': [],  # will be set dynamically using provisioner classes...
        'server': Url(),
        'shares': [{
            # The existence of the source directory will be checked!
            'source': IsDir(),
            'dest': str,
            'set_host_acl': bool,
        }],
        'shell': {
            'user': str,
            'home': str,
        },
        'users': [{
            # Usernames max length is set 32 characters according to useradd's man page.
            Required('name'): All(str, Length(max=32)),
            'home': str,
            'password': str,
        }],
    }

    def _check_provisioner_config(config):
        provisioners = Provisioner.provisioners.values()

        # Check if 'type' is correctly defined
        Schema({Required('type'): Any(*[provisioner.name for provisioner in provisioners])},
               extra=ALLOW_EXTRA)(config)

        # Check if the detected provisioner's schema is fully satisfied
        c = config.copy()
        name = c.pop('type')
        detected_provisioner = next(provisioner for provisioner in provisioners
                                    if provisioner.name == name)
        validated = Schema(detected_provisioner.schema)(c)
        validated['type'] = name
        return validated

    # Inserts provisioner specific schema rules in the global schema dict.
    _top_level_and_containers_common_options['provisioning'] = [All(_check_provisioner_config)]

    _container_options = {
        Required('name'): LXDIdentifier(),
    }
    _container_options.update(_top_level_and_containers_common_options)

    _lxdock_options = {
        Required('name'): LXDIdentifier(),
        'containers': [_container_options, ],
    }
    _lxdock_options.update(_top_level_and_containers_common_options)

    return Schema(_lxdock_options)


# The schema will be used to validate LXDock files!
schema = get_schema()



from .base import Guest


class AlpineGuest(Guest):
    """ This guest can provision Alpine containers. """

    name = 'alpine'

    def install_packages(self, packages):
        self.run(['apk', 'update'])
        self.run(['apk', 'add'] + packages)



import collections


def yesno(question, default=False):
    """ Asks the user to answer the yes/no question and returns the corresponding boolean. """
    question_suffix = ' [Yn] ' if default else ' [yN] '
    answer = input(question.strip() + question_suffix).strip().lower()
    answer_to_bool = collections.defaultdict(lambda: default)
    answer_to_bool.update({'yes': True, 'y': True, 'no': False, 'n': False, })
    return answer_to_bool[answer]



from lxdock.guests import DebianGuest
from lxdock.test import FakeContainer


class TestDebianGuest:
    def test_can_install_packages(self):
        guest = DebianGuest(FakeContainer())
        guest.install_packages(['python', 'openssh', ])
        assert guest.lxd_container.execute.call_count == 2
        assert guest.lxd_container.execute.call_args_list[0][0] == \
            (['apt-get', 'update', ], )
        assert guest.lxd_container.execute.call_args_list[1][0] == \
            (['apt-get', 'install', '-y', 'python', 'openssh', ], )



import os
import platform
import unittest.mock
from pathlib import Path

import pytest

from lxdock.hosts import Host
from lxdock.hosts.base import InvalidHost
from lxdock.test import FakeContainer


class TestGuest:
    def test_subclasses_cannot_be_registered_if_they_do_not_provide_a_guest_name(self):
        with pytest.raises(InvalidHost):
            class InvalidDummyHost(Host):
                name = None

    def test_can_detect_os_or_distribution(self):
        class DummyHost(Host):
            name = 'dummy'

        class PlatformHost(Host):
            name = platform.platform()

        assert not DummyHost.detect()
        assert PlatformHost.detect()

    @unittest.mock.patch.object(Path, 'open')
    def test_can_return_ssh_pubkey(self, mock_open):
        host = Host()
        assert host.get_ssh_pubkey()
        assert mock_open.call_count == 1

    @unittest.mock.patch('subprocess.Popen')
    def test_can_give_current_user_access_to_share(self, mocked_call):
        host = Host()
        host.give_current_user_access_to_share('.')
        assert mocked_call.call_count == 1
        assert mocked_call.call_args[0][0] == 'setfacl -Rdm u:{}:rwX .'.format(os.getuid())

    @unittest.mock.patch('subprocess.Popen')
    @unittest.mock.patch('os.stat')
    def test_can_give_mapped_user_access_to_share(self, mocked_stat, mocked_call):
        class MockedContainer(object):
            name = 'test'
        mocked_stat.return_value = unittest.mock.MagicMock(st_uid='19958953')
        host = Host()
        host.give_mapped_user_access_to_share(FakeContainer(), '.', '.')
        assert mocked_call.call_count == 1
        assert mocked_call.call_args[0] == (
            'setfacl -Rm user:lxd:rwx,default:user:lxd:rwx,user:19958953:rwx,default:user:19958953'
            ':rwx .',)



"""
    Base provisioner
    ================
    This module provides the `Provisioner` base class that is used to define the provisioners
    supported by LXDock (eg. Ansible, ...).
"""

import logging
import os

from ..utils.metaclass import with_metaclass


__all__ = ['Provisioner', ]

logger = logging.getLogger(__name__)


class InvalidProvisioner(Exception):
    """ The `Provisioner` subclass is not valid. """


class _ProvisionerBase(type):
    """ Metaclass for all LXDock provisioners.

    This metaclass ensures that all defined `Provisioner` subclasses have the required attributes
    and proceeds to some validation checks. Additionally it implements the "plugin mount" paradigm
    and stores a dictionary of `Provisioner` subclasses in the namespace of the "plugin mount"
    class (`Provisioner`).
    """

    def __new__(cls, name, bases, attrs):
        super_new = super(_ProvisionerBase, cls).__new__
        parents = [base for base in bases if isinstance(base, _ProvisionerBase)]

        # We stop here if we are considering the top-level class to which this metaclass was applied
        # and not one of its subclasses (eg. Provisioner).
        if not parents:
            return super_new(cls, name, bases, attrs)

        # Constructs the Provisioner class.
        new_provisioner = super_new(cls, name, bases, attrs)

        # Performs some validation checks.
        if not new_provisioner.name:
            raise InvalidProvisioner(
                "The 'name' attribute of Provisioner subclasses cannot be None")
        if not new_provisioner.schema:
            raise InvalidProvisioner(
                "The 'schema' attribute of Provisioner subclasses cannot be None")

        return new_provisioner

    def __init__(cls, name, bases, attrs):
        # We implement the "mount point" paradigm here in order to make all `Provisioner`
        # subclassses available in a single attribute.
        if not hasattr(cls, 'provisioners'):
            # The class has no provisioners dict: this means that we are considering the
            # "plugin mount" class. So we created the dict that will hold all the defined
            # `Provisioner` subclasses.
            cls.provisioners = {}
        else:
            # The `provisioners` attribute already exists so we are considering a `Provisioner`
            # subclass, which needs to be registered.
            cls.provisioners.update({cls.name.lower(): cls})


class Provisioner(with_metaclass(_ProvisionerBase)):
    """ Represents a single provisioner.

    `Provisioner` subclasses will be used by `Container` instances to run provisioning operations
    associated with the considered containers. For example they can be used to run Ansible playbooks
    to provision a web application on the container.
    """

    # The `name` of a provisioner is a required attribute and should always be set on `Provisioner`
    # subclasses.
    name = None

    # The `schema` of a provisioner should be a dictionary using voluptuous helpers. This dictionary
    # must define the "local" schema that should be provided in the LXDock file in order to use the
    # considered provisioner.
    schema = None

    # System packages needed for the provisioner to work correctly (on the guest side) should be
    # listed in attributed named "guest_required_packages_{guestname}". For example:
    #
    #     guest_required_packages_opensuse = ['openSSH', 'python3-base', ]
    #     guest_required_packages_oracle = ['openssh-server', 'python', ]
    #
    # These packages will be installed during the "provision" operation.

    def __init__(self, homedir, host, guests, options):
        self.homedir = homedir
        self.host = host
        self.guests = guests
        self.options = options.copy()

    def provision(self):
        """ Performs the provisioning operations using the considered provisioner. """
        # This method should be overriden in `Provisioner` subclasses that implement global
        # provisioning.
        self.setup()
        for guest in self.guests:
            self.provision_single(guest)

    def provision_single(self, guest):
        """ Performs the provisioning operations on the specified guest. """
        # This method should be overriden in `Provisioner` subclasses that don't implement
        # global provisioning.

    def setup(self):
        for guest in self.guests:
            if not guest.container.is_provisioned:
                self.setup_single(guest)

    def setup_single(self, guest):
        """ Setups the guest to run the provisioner if applicable. """
        # Ensure that the packages required to properly use the considered provisioner are installed
        # on the guest.
        logger.info('Preparing {0} up for provisioner {1}'.format(
            guest.container.name, self.name))
        required_packages = getattr(
            self, 'guest_required_packages_{}'.format(guest.name), None)
        if required_packages is not None:
            logger.info(
                'Installing packages required for the {} provisioner '
                'on the guest'.format(self.name))
            guest.install_packages(required_packages)

        # We allow `Provisioner` subclasses to define their own "setup_guest_{guestname}" methods
        # if setup operations have to be done on some specific guest prior to any provisioning
        # actions.
        guest_setup_method = getattr(self, 'setup_guest_{}'.format(guest.name), None)
        if guest_setup_method is not None:
            guest_setup_method(guest)

    ##################
    # HELPER METHODS #
    ##################

    def homedir_expanded_path(self, path):
        """ Expands the considered path with the path of the home homedir if applicable. """
        return os.path.join(self.homedir, path)



from lxdock.guests import OpenSUSEGuest
from lxdock.test import FakeContainer


class TestOpenSUSEGuest:
    def test_can_install_packages(self):
        guest = OpenSUSEGuest(FakeContainer())
        guest.install_packages(['python', 'openssh', ])
        assert guest.lxd_container.execute.call_count == 1
        assert guest.lxd_container.execute.call_args[0] == \
            (['zypper', '--non-interactive', 'install', 'python', 'openssh', ], )



from unittest.mock import Mock

from ..container import Container


class FakeContainer(Container):
    def __init__(self, project_name='project', homedir='/foo', client=None, **options):
        if client is None:
            client = Mock()
        options.setdefault('name', 'fakecontainer')
        super().__init__(project_name, homedir, client, **options)
        self._lxd_name = "{}-{}-123".format(project_name, self.name)

    def _get_container(self, create=True):
        result = Mock()
        result.execute.return_value = ('ok', 'ok', '')
        network_state = Mock()
        network_state.network = None
        result.state.return_value = network_state
        return result



import os

from lxdock.cli.project import get_project


FIXTURE_ROOT = os.path.join(os.path.dirname(__file__), 'fixtures')


def test_get_project_function_can_return_a_valid_project_instance():
    homedir = os.path.join(FIXTURE_ROOT, 'project01')
    project = get_project(homedir)
    assert project.name == 'project01'
    assert project.homedir == homedir



import os

from voluptuous import Any, Exclusive, IsFile

from .base import Provisioner


class ShellProvisioner(Provisioner):
    """ Allows to perform provisioning shell operations on the host/guest sides. """

    name = 'shell'
    schema = {
        Exclusive('inline', 'shelltype'): str,
        Exclusive('script', 'shelltype'): IsFile(),
        'side': Any('guest', 'host'),
    }

    def provision_single(self, guest):
        """ Executes the shell commands in the guest container or in the host. """
        if 'script' in self.options and self._is_for_guest:
            # First case: we have to run the script inside the container. So the first step is
            # to copy the content of the script to a temporary file in the container, ensure
            # that the script is executable and then run the script.
            guest_scriptpath = os.path.join('/tmp/', os.path.basename(self.options['script']))
            with open(self.homedir_expanded_path(self.options['script'])) as fd:
                guest.lxd_container.files.put(guest_scriptpath, fd.read())
            guest.run(['chmod', '+x', guest_scriptpath])
            guest.run([guest_scriptpath, ])
        elif 'script' in self.options and self._is_for_host:
            # Second case: the script is executed on the host side.
            self.host.run([self.homedir_expanded_path(self.options['script']), ])
        elif 'inline' in self.options:
            # Final case: we run a command directly inside the container or outside.
            host_or_guest = self.host if self._side == 'host' else guest
            host_or_guest.run(['sh', '-c', self.options['inline']])

    def setup(self):
        # nothing to set up, avoid spurious messages with this override.
        pass

    ##################################
    # PRIVATE METHODS AND PROPERTIES #
    ##################################

    @property
    def _is_for_guest(self):
        """ Returns True if this provisioner should run on the guest side. """
        return self._side == 'guest'

    @property
    def _is_for_host(self):
        """ Returns True if this provisioner should run on the host side. """
        return self._side == 'host'

    @property
    def _side(self):
        return self.options.get('side', 'guest')



from .base import Guest


class FedoraGuest(Guest):
    """ This guest can provision Fedora containers. """

    name = 'fedora'

    def install_packages(self, packages):
        self.run(['dnf', '-y', 'install', ] + packages)



import os
from test.support import EnvironmentVarGuard

import pytest

from lxdock.conf.config import Config
from lxdock.conf.exceptions import (ConfigFileInterpolationError, ConfigFileNotFoundError,
                                    ConfigFileValidationError)


FIXTURE_ROOT = os.path.join(os.path.dirname(__file__), 'fixtures')


class TestConfig:
    def test_can_be_initialized_from_a_base_dir(self):
        project_dir = os.path.join(FIXTURE_ROOT, 'project01')
        config = Config.from_base_dir(project_dir)
        assert config['name'] == 'project01'
        assert config['image'] == 'ubuntu/xenial'
        assert config['mode'] == 'pull'

    def test_works_if_multiple_config_files_are_found(self):
        project_dir = os.path.join(FIXTURE_ROOT, 'project_with_multiple_config_files')
        config = Config.from_base_dir(project_dir)
        assert config.homedir == project_dir
        assert config.filename == 'lxdock.yml'

    def test_can_find_a_config_file_from_a_project_subdirectory(self):
        project_dir = os.path.join(FIXTURE_ROOT, 'project01', 'subdir')
        config = Config.from_base_dir(project_dir)
        assert config.homedir == os.path.join(FIXTURE_ROOT, 'project01')
        assert config.filename == 'lxdock.yml'
        assert config['name'] == 'project01'
        assert config['image'] == 'ubuntu/xenial'
        assert config['mode'] == 'pull'

    def test_raises_an_error_if_the_config_file_cannot_be_found(self):
        project_dir = os.path.dirname(__file__)
        with pytest.raises(ConfigFileNotFoundError):
            Config.from_base_dir(project_dir)

    def test_raises_an_error_if_the_config_file_is_invalid(self):
        project_dir = os.path.join(FIXTURE_ROOT, 'project_with_invalid_config')
        with pytest.raises(ConfigFileValidationError):
            Config.from_base_dir(project_dir)

    def test_allows_to_get_items_from_config_dict(self):
        project_dir = os.path.join(FIXTURE_ROOT, 'project01')
        config = Config.from_base_dir(project_dir)
        assert config['name'] == 'project01'

    def test_allows_to_check_item_presence_in_config_dict(self):
        project_dir = os.path.join(FIXTURE_ROOT, 'project01')
        config = Config.from_base_dir(project_dir)
        assert 'name' in config

    def test_can_load_container_config_from_top_level_options_if_no_containers_are_defined(self):
        project_dir = os.path.join(FIXTURE_ROOT, 'project01')
        config = Config.from_base_dir(project_dir)
        assert config.containers == [
            {'image': 'ubuntu/xenial', 'mode': 'pull', 'name': 'default'},
        ]

    def test_can_properly_handle_configurations_defining_multiple_containers(self):
        project_dir = os.path.join(FIXTURE_ROOT, 'project02')
        config = Config.from_base_dir(project_dir)
        assert config.containers == [
            {'mode': 'pull', 'image': 'ubuntu/xenial', 'name': 'web'},
            {
                'mode': 'pull', 'hostnames': ['ci.local'], 'image': 'debian/jessie', 'name': 'ci',
                'privileged': True,
            },
        ]

    def test_can_serialize_the_parsed_config(self):
        project_dir = os.path.join(FIXTURE_ROOT, 'project01')
        config = Config.from_base_dir(project_dir)
        assert config.serialize() == 'image: ubuntu/xenial\nmode: pull\nname: project01\n'

    def test_can_read_lxc_config(self):
        project_dir = os.path.join(FIXTURE_ROOT, 'lxc_config')
        config = Config.from_base_dir(project_dir)
        assert config.containers[0]['lxc_config'] == {'global_key': 'global_value',
                                                      'cont1_key': 'cont1_value'}

    def test_can_manage_config_with_interpolated_variables_coming_from_environment_variables(self):
        project_dir = os.path.join(FIXTURE_ROOT, 'project_with_dynamic_variables')
        env = EnvironmentVarGuard()
        with env:
            env.set('DUMMY_VAR_01', 'test01')
            config = Config.from_base_dir(project_dir)
            assert config['provisioning'][0]['inline'] == 'touch /opt/test01'
            assert config['provisioning'][1]['inline'] == 'echo $thisisatest'

    def test_can_manage_config_with_interpolated_variables_coming_from_an_env_file(self):
        project_dir = os.path.join(FIXTURE_ROOT, 'project_with_dynamic_variables_and_env_file')
        config = Config.from_base_dir(project_dir)
        assert config['provisioning'][0]['inline'] == 'touch /opt/test01'
        assert config['provisioning'][1]['inline'] == 'echo $thisisatest'

    def test_raises_an_error_if_a_variable_cannot_be_substituded(self):
        project_dir = os.path.join(FIXTURE_ROOT, 'project_with_dynamic_variables')
        with pytest.raises(ConfigFileInterpolationError):
            Config.from_base_dir(project_dir)



import os

from lxdock.utils.identifier import folderid


def test_folderid_helper_returns_an_identifier_based_on_inode_numbers():
    path = '.'
    stats = os.stat(path)
    assert str(stats.st_ino) in folderid(path)



import logging
import tempfile

from voluptuous import Extra, IsFile, Required

from ..network import get_ip
from .base import Provisioner


logger = logging.getLogger(__name__)


class AnsibleProvisioner(Provisioner):
    """ Allows to perform provisioning operations using Ansible. """

    name = 'ansible'

    guest_required_packages_alpine = ['python', ]
    guest_required_packages_arch = ['python', ]
    guest_required_packages_centos = ['python', ]
    guest_required_packages_debian = ['apt-utils', 'aptitude', 'python', ]
    guest_required_packages_fedora = ['python2', ]
    guest_required_packages_gentoo = ['dev-lang/python', ]
    guest_required_packages_opensuse = ['python3-base', ]
    guest_required_packages_ol = ['python', ]
    guest_required_packages_ubuntu = ['apt-utils', 'aptitude', 'python', ]

    schema = {
        Required('playbook'): IsFile(),
        'ask_vault_pass': bool,
        'vault_password_file': IsFile(),
        'groups': {Extra: [str, ]},
        'lxd_transport': bool,
    }

    def get_inventory(self):
        def line(guest):
            if self.options.get('lxd_transport'):
                ip = guest.container.lxd_name
            else:
                ip = get_ip(guest.lxd_container)
            return '{} ansible_host={} ansible_user=root'.format(guest.container.name, ip)

        def fmtgroup(name, hosts):
            hosts = [host for host in hosts if host in guestnames]
            return '[{}]\n{}'.format(name, '\n'.join(hosts))

        all_hosts_lines = '\n'.join(line(guest) for guest in self.guests)
        groups = self.options.get('groups', {})
        guestnames = {guest.container.name for guest in self.guests}
        groups_lines = '\n\n'.join(fmtgroup(key, val) for key, val in groups.items())
        return '\n\n'.join([all_hosts_lines, groups_lines])

    def provision(self):
        """ Performs the provisioning operations using ansible-playbook. """
        self.setup()
        with tempfile.NamedTemporaryFile() as tmpinv:
            tmpinv.write(self.get_inventory().encode('ascii'))
            tmpinv.flush()
            self.host.run(self._build_ansible_playbook_command_args(tmpinv.name))

    def setup_single(self, guest):
        super().setup_single(guest)

        if self.options.get('lxd_transport'):
            # we don't need ssh
            return

        ssh_pkg_name = {
            'alpine': 'openssh',
            'arch': 'openssh',
            'gentoo': 'net-misc/openssh',
            'opensuse': 'openSSH',
        }.get(guest.name, 'openssh-server')
        guest.install_packages([ssh_pkg_name])

        # Make sure that sshd is started
        if guest.name == 'alpine':
            guest.run(['rc-update', 'add', 'sshd'])
            guest.run(['/etc/init.d/sshd', 'start'])
        elif guest.name in {'arch', 'centos', 'fedora'}:
            guest.run(['systemctl', 'enable', 'sshd'])
            guest.run(['systemctl', 'start', 'sshd'])
        elif guest.name == 'ol':
            guest.run(['/sbin/service', 'sshd', 'start'])

        # Add the current user's SSH pubkey to the container's root SSH config.
        ssh_pubkey = self.host.get_ssh_pubkey()
        if ssh_pubkey is not None:
            guest.add_ssh_pubkey_to_root_authorized_keys(ssh_pubkey)
        else:
            logger.warning('SSH pubkey was not found. Provisioning tools may not work correctly...')

    ##################################
    # PRIVATE METHODS AND PROPERTIES #
    ##################################

    def _build_ansible_playbook_command_args(self, inventory_filename):
        cmd_args = ['ANSIBLE_HOST_KEY_CHECKING=False', 'ansible-playbook', ]
        cmd_args.extend(['--inventory-file', inventory_filename, ])
        # Append the --ask-vault-pass option if applicable.
        if self.options.get('ask_vault_pass'):
            cmd_args.append('--ask-vault-pass')

        # Append the --vault-password-file option if applicable.
        vault_password_file = self.options.get('vault_password_file')
        if vault_password_file is not None:
            cmd_args.extend([
                '--vault-password-file', self.homedir_expanded_path(vault_password_file)])

        if self.options.get('lxd_transport'):
            cmd_args.extend(['-c', 'lxd', ])

        # Append the playbook filepath and return the final command.
        cmd_args.append(self.homedir_expanded_path(self.options['playbook']))
        return cmd_args



"""
    LXDock specific testcases & test wrappers
    =========================================
    This module provides testcase classes that can be used to perform common operations before or
    after test execution.
"""

from .. import constants
from ..client import get_client


__all__ = ['LXDTestCase', ]


TEST_CONTAINER_INJECTED_ID = 'lxdock-pytest-'


def _remove_test_containers(client=None):
    """ This function can be used to ensure test containers are removed after test execution. """
    client = client or get_client()
    test_containers = list(filter(
        lambda c: TEST_CONTAINER_INJECTED_ID in c.name, client.containers.all()))
    [c.stop(wait=True) for c in test_containers if c.status_code == constants.CONTAINER_RUNNING]
    [c.delete(wait=True) for c in test_containers]


class LXDTestCase:
    """ Base test class used to run integration tests that require the use of LXD. """

    @classmethod
    def teardown_class(cls):
        _remove_test_containers()

    def teardown_method(self, method):
        _remove_test_containers(client=self.client)

    ##################################
    # UTILITY METHODS AND PROPERTIES #
    ##################################

    def containername(self, name):
        """ Helper to construct a container name that should be deleted after each test. """
        return TEST_CONTAINER_INJECTED_ID + name

    @property
    def client(self):
        """ Returns a client instance to use when running tests requiring a LXD client. """
        if not hasattr(self, '_client'):
            self._client = get_client()
        return self._client



import logging

from . import constants
from .container import Container
from .exceptions import ProjectError
from .guests import Guest
from .hosts import Host
from .logging import (console_stderr_handler, console_stdout_handler, get_default_formatter,
                      get_per_container_formatter)
from .network import ContainerEtcHosts, EtcHosts
from .provisioners import Provisioner


logger = logging.getLogger(__name__)


class Project:
    """ A project is used to orchestrate a collection of containers. """

    def __init__(self, name, homedir, client, containers, provisioning_steps):
        self.name = name
        self.homedir = homedir
        self.client = client
        self.containers = containers
        self.provisioning_steps = provisioning_steps

    @classmethod
    def from_config(cls, project_name, client, config):
        """ Creates a `Project` instance from a config object. """
        containers = []
        for container_config in config.containers:
            containers.append(Container(project_name, config.homedir, client, **container_config))
        return cls(project_name, config.homedir, client, containers, config.provisioning_steps)

    #####################
    # CONTAINER ACTIONS #
    #####################

    def destroy(self, container_names=None):
        """ Destroys the containers of the project. """
        containers = [self.get_container_by_name(name) for name in container_names] \
            if container_names else self.containers
        for container in self._containers_generator(containers=containers):
            container.destroy()
        self._update_guest_etchosts()

    def halt(self, container_names=None):
        """ Stops containers of the project. """
        containers = [self.get_container_by_name(name) for name in container_names] \
            if container_names else self.containers
        for container in self._containers_generator(containers=containers):
            container.halt()
        self._update_guest_etchosts()

    def provision(self, container_names=None, force=True):
        """ Provisions the containers of the project. """
        # This happens in two phases: local provisioning, then global provisioning.
        # Local provisioning is the provisioning that is declared directly in the
        # `containers` config section. Global provisioning is declared at the root
        # of the config.
        containers = [self.get_container_by_name(name) for name in container_names] \
            if container_names else self.containers
        for container in self._containers_generator(containers=containers):
            container.provision(force=force)

        host = Host.get()
        if not force:
            containers = [c for c in containers if not c.is_provisioned]
        if not containers:
            return

        guests = [Guest.get(c) for c in containers]
        for provisioning_item in self.provisioning_steps:
            provisioning_type = provisioning_item['type'].lower()
            provisioner_class = Provisioner.provisioners.get(provisioning_type)
            if provisioner_class is not None:
                provisioner = provisioner_class(
                    self.homedir, host, guests, provisioning_item)
                logger.info('Global provisioning with {0}'.format(provisioner.name))
                provisioner.provision()

        for guest in guests:
            guest.lxd_container.config['user.lxdock.provisioned'] = 'true'
            guest.lxd_container.save(wait=True)

    def shell(self, container_name=None, **kwargs):
        """ Opens a new shell in our first container. """
        containers = [self.get_container_by_name(container_name)] if container_name \
            else self.containers
        if len(containers) > 1:
            raise ProjectError(
                'This action requires a container name to be specified because {count} '
                'containers are defined in this project.'.format(count=len(self.containers)))
        for container in self._containers_generator(containers=containers):
            container.shell(**kwargs)

    def status(self, container_names=None):
        """ Shows the statuses of the containers of the project. """
        containers = [self.get_container_by_name(name) for name in container_names] \
            if container_names else self.containers
        max_name_length = max(len(c.name) for c in containers)
        logger.info('Current container states:')
        for container in containers:
            logger.info('{container_name} ({status})'.format(
                container_name=container.name.ljust(max_name_length + 10), status=container.status))

    def up(self, container_names=None, provisioning_mode=None):
        """ Creates, starts and provisions the containers of the project. """
        containers = [self.get_container_by_name(name) for name in container_names] \
            if container_names else self.containers
        not_running = [c for c in containers if not c.is_running]
        if not not_running:
            logger.warning("Everything is already up and running! Nothing to do.")
            return
        [logger.info('Bringing container "{}" up'.format(c.name)) for c in not_running]
        for container in self._containers_generator(containers=not_running):
            container.up()
        self._update_guest_etchosts()

        # Provisions the container if applicable; that is only if it hasn't been provisioned before
        # or if the provisioning is manually enabled.
        provisioning_mode = provisioning_mode or constants.ProvisioningMode.AUTO
        if not provisioning_mode == constants.ProvisioningMode.DISABLED:
            force = provisioning_mode == constants.ProvisioningMode.ENABLED
            self.provision(container_names=[c.name for c in not_running], force=force)

    ##################################
    # UTILITY METHODS AND PROPERTIES #
    ##################################

    def get_container_by_name(self, name):
        """ Returns the `Container` instance associated with the given name. """
        containers_dict = {c.name: c for c in self.containers}
        if name in containers_dict:
            return containers_dict[name]

        # No containers exist for the considered name.
        raise ProjectError(
            'The container with the name "{name}" was not '
            'found for this project.'.format(name=name))

    ##################################
    # PRIVATE METHODS AND PROPERTIES #
    ##################################

    def _containers_generator(self, containers=None):
        containers = containers or self.containers
        for container in containers:
            console_stdout_handler.setFormatter(get_per_container_formatter(container.name))
            console_stderr_handler.setFormatter(get_per_container_formatter(container.name))
            yield container
        console_stdout_handler.setFormatter(get_default_formatter())
        console_stderr_handler.setFormatter(get_default_formatter())

    def _update_guest_etchosts(self):
        """ Updates /etc/hosts on **all** running lxdock-managed containers.

        ... even those outside the current project. This way, containers can contact themselves
        using the same domain names the host uses.
        """
        def should_update(c):
            return c.config.get('user.lxdock.made') and c.status_code == constants.CONTAINER_RUNNING
        # At this point, our host's /etc/hosts is fully updated. No need to go fetch IP's and stuff
        # we can just re-use what we've already computed in every container up/halt ops before.
        etchosts = EtcHosts()
        containers = (c for c in self.client.containers.all() if should_update(c))
        for container in containers:
            container_etchosts = ContainerEtcHosts(container)
            container_etchosts.lxdock_bindings = etchosts.lxdock_bindings
            container_etchosts.save()



# TEMPLATES
# --

INIT_LXDOCK_FILE_CONTENT = """# This is a basic configuration file for LXDock.
# All configuration is done through this YML file that should be placed at the root of your project.
# The file define a basic LXDock project containing a single container with highlights regarding
# some other useful options.
name: {project_name}
image: {image}

# By default LXDock creates a single "default" container if you don't specify a "containers" option.
# But you need the "containers" option if you have more than one container.

# containers:
#   - name: {project_name}01
#   - name: {project_name}02
#   - name: {project_name}03
#     Most of the options can be redefined for each container definition, eg. the "image" option:
#     image: archlinux

# You can use the "provisioning" option to define provisioning tools that should be used to
# provision your containers. For example, you could use Ansible as follows:
# provisioning:
#   - type: ansible
#     playbook: deploy/site.yml

# A common need is to access your project folder in your containers. To do this you can use the
# "shares" option:
# shares:
#   - source: .
#     dest: /myshare
"""



import codecs
import sys
from os.path import abspath
from os.path import dirname
from os.path import join

from setuptools import setup, find_packages

import lxdock


if sys.version_info < (3, 4):
    sys.exit('lxdock requires Python 3.4+ to run')


def read_relative_file(filename):
    """ Returns contents of the given file, whose path is supposed relative to this module. """
    with codecs.open(join(dirname(abspath(__file__)), filename), encoding='utf-8') as f:
        return f.read()


setup(
    name='lxdock',
    version=lxdock.__version__,
    author='Virgil Dupras, Morgan Aubert',
    author_email='',
    packages=find_packages(exclude=['tests.*', 'tests']),
    include_package_data=True,
    url='https://github.com/lxdock/lxdock',
    license='GPLv3',
    description='LXDock is a tool for orchestrating LXD containers.',
    long_description=read_relative_file('README.rst'),
    keywords='lxd lxc containers environments orchestration devops',
    zip_safe=False,
    install_requires=[
        'colorlog>=2.0,<3.0',
        'pylxd>=2.2.4',
        'python-dotenv>=0.6',
        'PyYAML>=3.0,<4.0',
        'voluptuous>=0.9,<1.0',
    ],
    tests_require=[
        'pytest',
    ],
    entry_points={
        'console_scripts': [
            'lxdock = lxdock.cli.main:main',
        ],
    },
    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Console',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Natural Language :: English',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
    ],
)



from lxdock.guests import ArchLinuxGuest
from lxdock.test import FakeContainer


class TestArchLinuxGuest:
    def test_can_install_packages(self):
        guest = ArchLinuxGuest(FakeContainer())
        guest.install_packages(['python', 'openssh', ])
        assert guest.lxd_container.execute.call_count == 1
        assert guest.lxd_container.execute.call_args[0] == \
            (['pacman', '-S', '--noconfirm', 'python', 'openssh'], )



"""
    Base guest
    ==========
    This module provides the `Guest` base class that is used to define the guest OS or distributions
    supported by LXDock (eg. debian, ...).
"""

import logging
import re
import tarfile
import tempfile
from pathlib import Path, PurePosixPath

from pylxd.exceptions import NotFound

from ..utils.metaclass import with_metaclass


__all__ = ['Guest', ]

logger = logging.getLogger(__name__)


class InvalidGuest(Exception):
    """ The `Guest` subclass is not valid. """


class _GuestBase(type):
    """ Metaclass for all LXD guests.

    This metaclass ensures that all defined `Guest` subclasses have the required attributes and
    proceeds to some validation checks. Additionally it implements the "plugin mount" paradigm and
    stores a list of `Guest` subclasses in the namespace of the "plugin mount" class (`Guest`).
    """

    def __new__(cls, name, bases, attrs):
        super_new = super(_GuestBase, cls).__new__
        parents = [base for base in bases if isinstance(base, _GuestBase)]

        # We stop here if we are considering the top-level class to which this metaclass was applied
        # and not one of its subclasses (eg. Guest).
        if not parents:
            return super_new(cls, name, bases, attrs)

        # Constructs the Guest class.
        new_guest = super_new(cls, name, bases, attrs)

        # Performs some validation checks.
        if not new_guest.name:
            raise InvalidGuest("The 'name' attribute of Guest subclasses cannot be None")

        return new_guest

    def __init__(cls, name, bases, attrs):
        # We implement the "mount point" paradigm here in order to make all `Guest` subclassses
        # available in a single attribute.
        if not hasattr(cls, 'guests'):
            # The class has no guests list: this means that we are considering the "plugin mount"
            # class. So we created the list that will hold all the defined `Guest` subclasses.
            cls.guests = []
        else:
            # The `guests` attribute already exists so we are considering a `Guest` subclass, which
            # needs to be registered.
            cls.guests.append(cls)


class Guest(with_metaclass(_GuestBase)):
    """ Represents a single guest.

    `Guest` subclasses will be used by `Container` instances to perform common operations on the
    guest side. For example they can be used to ensure that some packages are installed when doing
    barebone setups, to create some users, ... `Guest` subclasses should correspond to specific OSes
    or distributions that can be runned as LXD containers.
    """

    # The `name` of a guest is a required attribute and should always be set on `Guest` subclasses.
    name = None

    def __init__(self, container):
        self.container = container
        self.lxd_container = container._container

    @classmethod
    def detect(cls, lxd_container):
        """ Detects if the container is an "instance" of the considered OS.

        The operations defined in this method assume that OS/distribution information will be
        available through the analysis of files such as ""/etc/os-release", ""/etc/lsb_release" or
        ""/etc/issue". This classmethod should be overriden if these methods cannot be applied to
        the considered OS/distribution.
        """
        candidates = [
            ('/etc/os-release', '(id|ID)="?{}"?'.format(cls.name.lower())),
            ('/etc/lsb_release', '(id|ID)="?{}"?'.format(cls.name.lower())),
            ('/etc/issue', cls.name.lower()),
        ]
        found = False
        for candidate, pattern in candidates:
            try:
                candidate_filecontent = lxd_container.files.get(candidate)
            except NotFound:
                continue
            found = len(re.findall(pattern, str(candidate_filecontent.lower()))) > 0
            if found:
                break
        return found

    @classmethod
    def get(cls, container):
        """ Returns the `Guest` instance associated with the considered container. """
        class_ = next((k for k in cls.guests if k.detect(container._container)), Guest)
        return class_(container)

    def add_ssh_pubkey_to_root_authorized_keys(self, pubkey):
        """ Add a given SSH public key to the root user's authorized keys. """
        logger.info("Adding {} to machine's authorized keys".format(pubkey))
        self.run(['mkdir', '-p', '/root/.ssh'])
        self.lxd_container.files.put('/root/.ssh/authorized_keys', pubkey)

    def create_user(self, username, home=None, password=None):
        """ Adds the passed user to the container system. """
        options = ['--create-home', ]
        if home is not None:
            options += ['--home-dir', home, ]
        if password is not None:
            options += ['-p', password, ]
        self.run(['useradd', ] + options + [username, ])

    ########################################################
    # METHODS THAT SHOULD BE OVERRIDEN IN GUEST SUBCLASSES #
    ########################################################

    def install_packages(self, packages):  # pragma: no cover
        """ Installs the considered packages on the guest.

        :param packages: The list of packages to install on the guest
        :type packages: list
        """
        # This method should be overriden in `Guest` subclasses.
        self._warn_guest_not_supported('for installing packages')

    ##################
    # HELPER METHODS #
    ##################

    def run(self, cmd_args):
        """ Runs the specified command inside the current container. """
        logger.debug('Running {0}'.format(' '.join(cmd_args)))
        exit_code, stdout, stderr = self.lxd_container.execute(cmd_args)
        logger.debug(stdout)
        logger.debug(stderr)
        return exit_code

    def copy_file(self, host_path, guest_path):
        """
        Copies a file from host_path (pathlib.Path) to guest_path (pathlib.PurePath).
        Ensures `mkdir -p` before calling LXD file put.
        """
        self.run(['mkdir', '-p', str(guest_path.parent)])
        with host_path.open('rb') as f:
            logger.debug('Copying host:{} to guest:{}'.format(host_path, guest_path))
            self.lxd_container.files.put(str(guest_path), f.read())

    _guest_temporary_tar_path = '/.lxdock.d/copied_directory.tar'

    def copy_directory(self, host_path, guest_path):
        """
        Copies a directory from host_path (pathlib.Path) to guest_path (pathlib.PurePath).
        This is natively supported since LXD 2.2 but we have to support 2.0+
        Refs: https://github.com/lxc/lxd/issues/2401

        Uses tar to pack/unpack the directory.
        """
        guest_tar_path = self._guest_temporary_tar_path
        self.run(['mkdir', '-p', str(guest_path)])
        with tempfile.NamedTemporaryFile() as f:
            logger.debug("Creating tar file from {}".format(host_path))
            tar = tarfile.open(f.name, 'w')
            tar.add(str(host_path), arcname='.')
            tar.close()
            self.copy_file(Path(f.name), PurePosixPath(guest_tar_path))
        self.run(['tar', '-xf', guest_tar_path, '-C', str(guest_path)])
        self.run(['rm', '-f', str(guest_tar_path)])

    ##################################
    # PRIVATE METHODS AND PROPERTIES #
    ##################################

    def _warn_guest_not_supported(self, for_msg):  # pragma: no cover
        """ Warns the user that a specific operation cannot be performed. """
        logger.warn('Guest not supported {}, doing nothing...'.format(for_msg))



import pytest
from voluptuous.error import ValueInvalid

from lxdock.conf.validators import Hostname, LXDIdentifier


class TestHostnameValidator:
    def test_can_validate_a_basic_hostname(self):
        hostame_validator = Hostname('validator')
        assert hostame_validator('myhost')

    def test_can_validate_a_dotted_hostname(self):
        hostame_validator = Hostname('validator')
        assert hostame_validator('myhost.domain')

    def test_can_validate_a_dotted_hostname_ending_with_a_dot(self):
        hostame_validator = Hostname('validator')
        assert hostame_validator('myhost.domain.')

    def test_cannot_validate_a_hostname_taht_is_longer_than_255_characters(self):
        hostame_validator = Hostname('validator')
        with pytest.raises(ValueInvalid):
            hostame_validator('bad' * 100)


class TestLXDIdentifier:
    def test_can_validate_a_basic_identifier(self):
        id_validator = LXDIdentifier('validator')
        assert id_validator('myidentifier')

    def test_can_validate_identifiers_made_up_of_letters_digits_and_dashes(self):
        id_validator = LXDIdentifier('validator')
        assert id_validator('myidentifier01')
        assert id_validator('myidentifier-02-test')

    def test_cannot_allow_identifiers_that_do_not_start_with_letters(self):
        id_validator = LXDIdentifier('validator')
        with pytest.raises(ValueInvalid):
            id_validator('42identifier')

    def test_cannot_allow_identifiers_longer_than_63_characters(self):
        id_validator = LXDIdentifier('validator')
        assert id_validator('i' * 63)
        with pytest.raises(ValueInvalid):
            id_validator('i' * 64)
